package pt.uc.dei.crf_ma;

import java.io.File;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.io.PrintWriter;
import java.text.DecimalFormat;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.HashSet;
import java.util.TreeMap;

import pt.uc.dei.crf_ma.types.TermExtractorInstance;
import cc.mallet.pipe.SerialPipes;
import cc.mallet.types.Instance;
import cc.mallet.types.InstanceList;

public class CRFWithMultipleTaggers {

//	public static final String[] START_TAGS = {"I-KEYWORD"};
//	public static final String[] CONTINUE_TAGS = {"C-KEYWORD"};
	public static final String[] START_TAGS = {"B-LOC", "B-MISC", "B-ORG", "B-PER"};
	public static final String[] CONTINUE_TAGS = {"I-LOC", "I-MISC", "I-ORG", "I-PER"};
	//public static final String[] START_TAGS = {"B-NP"};
    //public static final String[] CONTINUE_TAGS = {"I-NP"};
	
	public static PrintWriter logPW;
	
	private int maxIterations;
	private int nthreads;
	private String inputFolder; // folder holding the input data
	private String groundTruthFolder; // folder holding the ground truth for the input data (when available)
	private String evaluationFolder; // folder holding an independent evaluation dataset (when available)
	private SequenceLabeller sequenceLabeller; // the CRF model
	private int totalDataInstances = 0; // holds the total number of data instances for training in the input training data
	private TreeMap<String, TreeMap<String, InstanceList>> taggersDataInstances; // holds the instances of the files of each tagger
	private TreeMap<String, ArrayList<String>> taggersDataFiles; // holds the names of the files of each tagger
	private TreeMap<String, InstanceList> distinctInputInstances; // holds all the distinct instances found in the input folder, to determine ground truth, measure accuracy, sensitivity, etc.
	private TreeMap<String, InstanceList> actualGroundTruth; // holds the actual ground truth of the data (when available)
	private TreeMap<String, InstanceList> majoratyVotingGroundTruth1; // holds the estimation of the ground truth using majority voting
	private TreeMap<String, InstanceList> majoratyVotingGroundTruth2; // holds the estimation of the ground truth using majority voting
	private TreeMap<String, InstanceList> majoratyVotingGroundTruth3; // holds the estimation of the ground truth using majority voting
	private TreeMap<String, InstanceList> evaluationData; // holds the evaluation data
	private MTTrainer mtTrainer;

	public CRFWithMultipleTaggers(String inputFolder, String groundTruthFolder, String evaluationFolder, int maxIterations, int nthreads) {
		System.out.println("[MultipleTaggers] Initializing... (new maven version)");
		System.out.flush();
	    
		this.maxIterations = maxIterations;
		this.nthreads = nthreads;
		this.inputFolder = inputFolder;
		if (!this.inputFolder.endsWith("/")) {
			this.inputFolder += "/";
		}
		this.groundTruthFolder = groundTruthFolder;
		if (!this.groundTruthFolder.endsWith("/")) {
			this.groundTruthFolder += "/";
		}
		this.evaluationFolder = evaluationFolder;
		if (evaluationFolder != null && !this.evaluationFolder.endsWith("/")) {
			this.evaluationFolder += "/";
		}
		
		// initialize term extractor
		initializeTermExtractor();

		//  load ground truth of the data
		this.actualGroundTruth = loadActualGroundTruth(this.groundTruthFolder);
		
		// load evaluation data
		if (evaluationFolder != null) {
			this.evaluationData = loadEvaluationData(this.evaluationFolder); 
		}
		
		// load taggers training data files
		this.totalDataInstances = 0; 
		this.taggersDataFiles = getInputFilesFromFolder(this.inputFolder);
		this.taggersDataInstances = loadDataInstences(this.taggersDataFiles);
		System.out.flush();
		
		// estimate ground truth using majority voting
		this.majoratyVotingGroundTruth1 = estimateGroundTruthUsingMajorityVoting1(taggersDataInstances, distinctInputInstances);
		this.majoratyVotingGroundTruth2 = estimateGroundTruthUsingMajorityVoting2(taggersDataInstances, distinctInputInstances);
		this.majoratyVotingGroundTruth3 = estimateGroundTruthUsingMajorityVoting3(taggersDataInstances, distinctInputInstances);
		System.out.flush();
		
		// evaluate estimate ground truth using majority voting
		DecimalFormat f = new DecimalFormat("0.####");
		System.out.println("[MultipleTaggers] Evaluating majority voting ground truth...");
		EvaluationResults evaluationResults1 = CRFWithMultipleTaggers.evaluateAgainstGroundTruth(majoratyVotingGroundTruth1, actualGroundTruth);
		System.out.println("[MultipleTaggers] \tResults: tokenAcc="+f.format(evaluationResults1.getTokenAccuracy())+
				" precision="+f.format(evaluationResults1.getPrecision())+
				" recall="+f.format(evaluationResults1.getRecall())+" f1="+f.format(evaluationResults1.getF1()));
		System.out.println("[MultipleTaggers] Evaluating majority voting (v2) ground truth...");
		EvaluationResults evaluationResults2 = CRFWithMultipleTaggers.evaluateAgainstGroundTruth(majoratyVotingGroundTruth2, actualGroundTruth);
		System.out.println("[MultipleTaggers] \tResults: tokenAcc="+f.format(evaluationResults2.getTokenAccuracy())+
				" precision="+f.format(evaluationResults2.getPrecision())+
				" recall="+f.format(evaluationResults2.getRecall())+" f1="+f.format(evaluationResults2.getF1()));
		System.out.println("[MultipleTaggers] Evaluating majority voting (v3) ground truth...");
		EvaluationResults evaluationResults3 = CRFWithMultipleTaggers.evaluateAgainstGroundTruth(majoratyVotingGroundTruth3, actualGroundTruth);
		System.out.println("[MultipleTaggers] \tResults: tokenAcc="+f.format(evaluationResults3.getTokenAccuracy())+
				" precision="+f.format(evaluationResults3.getPrecision())+
				" recall="+f.format(evaluationResults3.getRecall())+" f1="+f.format(evaluationResults3.getF1()));
		if (logPW != null) {
			logPW.println("MV1 (train): " + f.format(evaluationResults1.getPrecision()) + "\t" + f.format(evaluationResults1.getRecall()) + "\t" + f.format(evaluationResults1.getF1()));
			logPW.println("MV2 (train): " + f.format(evaluationResults2.getPrecision()) + "\t" + f.format(evaluationResults2.getRecall()) + "\t" + f.format(evaluationResults2.getF1()));
			logPW.println("MV3 (train): " + f.format(evaluationResults3.getPrecision()) + "\t" + f.format(evaluationResults3.getRecall()) + "\t" + f.format(evaluationResults3.getF1()));
			logPW.flush();
		}
		System.out.flush();
		
		// compute annotators' true F1s
		String f1s = "";
		for (String taggerName : taggersDataInstances.keySet()) {
			TreeMap<String, InstanceList> taggerInstances = taggersDataInstances.get(taggerName);
			EvaluationResults evaluationResults = CRFWithMultipleTaggers.evaluateAgainstGroundTruth(taggerInstances, actualGroundTruth);
			f1s += f.format(evaluationResults.getF1()) + ",";
		}
		System.out.println("[MultipleTaggers] Annotators: (true) F1s: " + f1s);
		System.out.flush();
		if (logPW != null) {
			logPW.println("Annotators (true) F1s: " + f1s);
			logPW.flush();
		}
		
		// train CRF with MV1 GT estimation and test it on test data
		System.out.flush();
		TreeMap<String, Double> trainingPercentages = new TreeMap<String, Double>();
		trainingPercentages.put("MV1", 1.0);
		TreeMap<String, TreeMap<String, InstanceList>> trainingData = new TreeMap<String, TreeMap<String, InstanceList>>();
		trainingData.put("MV1", majoratyVotingGroundTruth1);
		EvaluationResults[] evaluationResults = trainCRF(trainingData, trainingPercentages);
		EvaluationResults gtEvaluationResults = evaluationResults[0];
		EvaluationResults testDataEvaluationResults = evaluationResults[1];
		if (logPW != null) {
			logPW.println("MV1 CRF (train): " + f.format(gtEvaluationResults.getPrecision()) + "\t" + f.format(gtEvaluationResults.getRecall()) + "\t" + f.format(gtEvaluationResults.getF1()));
			if (testDataEvaluationResults != null) {
				logPW.println("MV1 CRF (test): " + f.format(testDataEvaluationResults.getPrecision()) + "\t" + f.format(testDataEvaluationResults.getRecall()) + "\t" + f.format(testDataEvaluationResults.getF1()));
			}
			logPW.flush();
		}
		
		// train CRF with MV2 GT estimation and test it on test data
		System.out.flush();
		trainingPercentages = new TreeMap<String, Double>();
		trainingPercentages.put("MV2", 1.0);
		trainingData = new TreeMap<String, TreeMap<String, InstanceList>>();
		trainingData.put("MV2", majoratyVotingGroundTruth2);
		evaluationResults = trainCRF(trainingData, trainingPercentages);
		gtEvaluationResults = evaluationResults[0];
		testDataEvaluationResults = evaluationResults[1];
		if (logPW != null) {
			logPW.println("MV2 CRF (train): " + f.format(gtEvaluationResults.getPrecision()) + "\t" + f.format(gtEvaluationResults.getRecall()) + "\t" + f.format(gtEvaluationResults.getF1()));
			if (testDataEvaluationResults != null) {
				logPW.println("MV2 CRF (test): " + f.format(testDataEvaluationResults.getPrecision()) + "\t" + f.format(testDataEvaluationResults.getRecall()) + "\t" + f.format(testDataEvaluationResults.getF1()));
			}
			logPW.flush();
		}
		
		
		// train CRF with MV3 GT estimation and test it on test data
		System.out.flush();
		trainingPercentages = new TreeMap<String, Double>();
		trainingPercentages.put("MV3", 1.0);
		trainingData = new TreeMap<String, TreeMap<String, InstanceList>>();
		trainingData.put("MV3", majoratyVotingGroundTruth3);
		evaluationResults = trainCRF(trainingData, trainingPercentages);
		gtEvaluationResults = evaluationResults[0];
		testDataEvaluationResults = evaluationResults[1];
		if (logPW != null) {
			logPW.println("MV3 CRF (train): " + f.format(gtEvaluationResults.getPrecision()) + "\t" + f.format(gtEvaluationResults.getRecall()) + "\t" + f.format(gtEvaluationResults.getF1()));
			if (testDataEvaluationResults != null) {
				logPW.println("MV3 CRF (test): " + f.format(testDataEvaluationResults.getPrecision()) + "\t" + f.format(testDataEvaluationResults.getRecall()) + "\t" + f.format(testDataEvaluationResults.getF1()));
			}
			logPW.flush();
		}
	}
	
	public void initializeTermExtractor() {
		SerialPipes featurePipeline = null;
		if (this.sequenceLabeller != null) {
			featurePipeline = this.sequenceLabeller.getFeaturePipeline();
		}
		
		// initialize term extractor
		this.sequenceLabeller = new SequenceLabeller();
		
		// set the special options to the TermExtractor
		this.sequenceLabeller.setEvalPeriod(100);
		this.sequenceLabeller.setNFolds(0); // no cross-validation
		this.sequenceLabeller.setTrainingFraction(1.0);
		this.sequenceLabeller.setRandomSeed(0);
		// this.sequenceLabeller.setDebugPipes(true);
		
		if (featurePipeline != null) {
			this.sequenceLabeller.setFeaturePipeline(featurePipeline);
		}
	}

	private TreeMap<String, ArrayList<String>> getInputFilesFromFolder(String inputFolder) {
		System.out.println("[MultipleTaggers]- Getting taggers filenames from folder: " + inputFolder);
		TreeMap<String, ArrayList<String>> taggersDataFiles = new TreeMap<String, ArrayList<String>>();

		File dir = new File(inputFolder);
		String[] children = dir.list();
		if (children == null) {
			throw new RuntimeException("Data directory not found: " + inputFolder);
		} else {
			for (int i = 0; i < children.length; i++) {
				String taggerName = children[i];
				if (!taggerName.equals(".DS_Store") && !taggerName.equals("._.DS_Store")) {
					System.out.println("[MultipleTaggers]- 	New tagger: " + taggerName);
					ArrayList<String> inputFiles = new ArrayList<String>();
					File subdir = new File(inputFolder + "" + taggerName + "/");
					String[] subchildren = subdir.list();
					if (subchildren == null) {
						throw new RuntimeException("Data directory not found: " + subdir.getAbsolutePath());
					} else {
						for (int j = 0; j < subchildren.length; j++) {
							String inputFilename = subchildren[j];
							if (!inputFilename.equals(".DS_Store") && !inputFilename.equals("._.DS_Store")) {
								//System.out.println("[MultipleTaggers]- 		New inputfile: " + inputFilename);
								inputFiles.add(inputFilename);
							}
						}
					}
					taggersDataFiles.put(taggerName, inputFiles);
				}
			}
		}
		System.out.println("[MultipleTaggers] Number of taggers: " + taggersDataFiles.size());
		return taggersDataFiles;
	}

	private TreeMap<String, TreeMap<String, InstanceList>> loadDataInstences(TreeMap<String, ArrayList<String>> taggersDataFiles) {
		TreeMap<String, TreeMap<String, InstanceList>> taggersDataInstances = new TreeMap<String, TreeMap<String, InstanceList>>();
		if (distinctInputInstances == null) {
			distinctInputInstances = new TreeMap<String, InstanceList>();
		}
		int i = 0;
		for (String taggerName : taggersDataFiles.keySet()) {
			i++;
			System.out.println("[MultipleTaggers] Loading instances for tagger: " + taggerName + " ["+i+"/"+taggersDataFiles.keySet().size()+"]");
			TreeMap<String, InstanceList> taggerInstances = sequenceLabeller.loadInstancesFromFolder(inputFolder + "" + taggerName, distinctInputInstances);
			System.out.println("[MultipleTaggers] 	Number of files loaded: " + taggersDataFiles.get(taggerName).size());
			int numInstances = 0;
			for (String filename : taggerInstances.keySet()) {
				numInstances += taggerInstances.get(filename).size();
			}
			System.out.println("[MultipleTaggers] 	Number of instances loaded: " + numInstances);
			totalDataInstances += numInstances;
			taggersDataInstances.put(taggerName, taggerInstances);	
		}
		System.out.println("[MultipleTaggers] Distinct instances found: (total="+distinctInputInstances.size()+")");
		for (String instanceId : distinctInputInstances.keySet()) {
			System.out.println("[MultipleTaggers] 	" + instanceId);
		}
		return taggersDataInstances;
	}
	
	private TreeMap<String, InstanceList> loadActualGroundTruth(String inputFolder) {
		System.out.println("[MultipleTaggers] Loading ground truth from folder: " + inputFolder);
		TreeMap<String, InstanceList> actualGroundTruth = sequenceLabeller.loadInstancesFromFolder(inputFolder, null);
		return actualGroundTruth;
	}
	
	private TreeMap<String, InstanceList> loadEvaluationData(String inputFolder) {
		System.out.println("[MultipleTaggers] Loading evaluation data from folder: " + inputFolder);
		TreeMap<String, InstanceList> evaluationData = sequenceLabeller.loadInstancesFromFolder(inputFolder, null);
		return evaluationData;
	}
	
	private TreeMap<String, InstanceList> estimateGroundTruthUsingMajorityVoting3(TreeMap<String, TreeMap<String, InstanceList>> taggersDataInstances, TreeMap<String, InstanceList> distinctInputInstances) {
		System.out.println("[MultipleTaggers] Estimating ground truth using majority voting (v3)...");
		TreeMap<String, InstanceList> majoratyVotingGroundTruth = new TreeMap<String, InstanceList>();
		for (String filename : distinctInputInstances.keySet()) {
			InstanceList instances = distinctInputInstances.get(filename);
			
			// iterate over instances and determine majority voting
			InstanceList mvInstances = new InstanceList(sequenceLabeller.getFeaturePipeline());
			for (int i=0; i<instances.size(); i++) {
				TermExtractorInstance instance = (TermExtractorInstance) instances.get(i);
				StringBuffer sb = new StringBuffer();
				
				// majority voting - get votes
				ArrayList<HashMap<String, Integer>> segCountsList = new ArrayList<HashMap<String, Integer>>();
				ArrayList<HashMap<String, Integer>> labelCountsList = new ArrayList<HashMap<String, Integer>>();
				ArrayList<String> mvTokens = instance.getTokens();
				for (int j=0; j<mvTokens.size(); j++) {
					segCountsList.add(new HashMap<String, Integer>());
					labelCountsList.add(new HashMap<String, Integer>());
				}
				for (String taggerName : taggersDataInstances.keySet()) {
					InstanceList taggerInstances = taggersDataInstances.get(taggerName).get(filename);
					if (taggerInstances == null || instances.size() != taggerInstances.size()) {
						continue; // skip this tagger for this filename
					}
					TermExtractorInstance taggerInstance = (TermExtractorInstance) taggerInstances.get(i);
					ArrayList<String> taggerLabels = taggerInstance.getLabels();
					if (mvTokens.size() != taggerLabels.size()) {
						continue; // skip this tagger for this instance
					}
					for (int j=0; j<mvTokens.size(); j++) {
						String taggerSegLabel = null;
						try {
							taggerSegLabel = taggerLabels.get(j).substring(0,1);
						} catch(Exception e) {
							continue;
						}
						String taggerLabel = taggerLabels.get(j).substring(1);
						//System.out.println("[MultipleTaggers] taggerSegLabel="+taggerSegLabel+" taggerLabel="+taggerLabel);
						if (!segCountsList.get(j).containsKey(taggerSegLabel)) {
							segCountsList.get(j).put(taggerSegLabel, 0);
						}
						segCountsList.get(j).put(taggerSegLabel, segCountsList.get(j).get(taggerSegLabel) + 1);
						if (!labelCountsList.get(j).containsKey(taggerLabel)) {
							labelCountsList.get(j).put(taggerLabel, 0);
						}
						labelCountsList.get(j).put(taggerLabel, labelCountsList.get(j).get(taggerLabel) + 1);
					}
				}
				
				// majority voting - get most votes labels
				ArrayList<String> mvLabels = new ArrayList<String>();
				String labels = "";
				for (int j=0; j<segCountsList.size(); j++) {
					HashMap<String, Integer> segCounts = segCountsList.get(j);
					int highestSegCount = -1;
					String highestSegLabel = null;
					for (String segLabel : segCounts.keySet()) {
						int count = segCounts.get(segLabel);
						if (count > highestSegCount) {
							highestSegCount = count;
							highestSegLabel = segLabel;
						}
					}
					if (highestSegLabel.equals("O")) {
						mvLabels.add("O");
						labels += "O\n";
						sb.append(instance.getTokens().get(j)+" O\n");
					} else {
						HashMap<String, Integer> labelCounts = labelCountsList.get(j);
						int highestLabelCount = -1;
						String highestLabel = null;
						for (String label : labelCounts.keySet()) {
							int count = labelCounts.get(label);
							if (count > highestLabelCount) {
								highestLabelCount = count;
								highestLabel = label;
							}
						}
						mvLabels.add(highestSegLabel+""+highestLabel);
						labels += highestSegLabel+""+highestLabel+"\n";
						sb.append(instance.getTokens().get(j)+" "+highestSegLabel+""+highestLabel+"\n");
					}
				}
				
				Instance mvInstanceTemp = new Instance(sb.toString(), null, instance.getName(), null);
				mvInstances.addThruPipe(mvInstanceTemp);
				mvInstanceTemp = mvInstances.remove(mvInstances.size()-1);
				TermExtractorInstance mvInstance = new TermExtractorInstance(mvInstanceTemp.getData(), mvInstanceTemp.getTarget(), mvInstanceTemp.getName(), mvInstanceTemp.getSource());
				mvInstance.setInstanceId(instance.getInstanceId());
				mvInstance.setTokens(instance.getTokens());
				mvInstance.setLabels(mvLabels);
				mvInstances.add(mvInstance);
			}
			majoratyVotingGroundTruth.put(filename, mvInstances);
		}
		
		PrintWriter pwMV = null;
		try {
			pwMV = new PrintWriter("majority_vote_ground_truth3.txt");
		} catch (FileNotFoundException e) {
			e.printStackTrace();
		}
		for (String filename : majoratyVotingGroundTruth.keySet()) {
			InstanceList instances = majoratyVotingGroundTruth.get(filename);
			for (int i=0; i<instances.size(); i++) {
				TermExtractorInstance instance = (TermExtractorInstance) instances.get(i);
				ArrayList<String> tokens = instance.getTokens();
				ArrayList<String> labels = instance.getLabels();
				for (int j=0; j<tokens.size(); j++) {
					pwMV.println(tokens.get(j) + "\t" + labels.get(j));
				}
				pwMV.println();
				pwMV.flush();
			}
		}
		pwMV.flush();
		pwMV.close();
		
		return majoratyVotingGroundTruth;
	}
	
	private TreeMap<String, InstanceList> estimateGroundTruthUsingMajorityVoting2(TreeMap<String, TreeMap<String, InstanceList>> taggersDataInstances, TreeMap<String, InstanceList> distinctInputInstances) {
		System.out.println("[MultipleTaggers] Estimating ground truth using majority voting (v2)...");
		TreeMap<String, InstanceList> majoratyVotingGroundTruth = new TreeMap<String, InstanceList>();
		for (String filename : distinctInputInstances.keySet()) {
			InstanceList instances = distinctInputInstances.get(filename);
			
			// iterate over instances and determine majority voting
			InstanceList mvInstances = new InstanceList(sequenceLabeller.getFeaturePipeline());
			for (int i=0; i<instances.size(); i++) {
				TermExtractorInstance instance = (TermExtractorInstance) instances.get(i);
				StringBuffer sb = new StringBuffer();
				
				// majority voting - get votes
				ArrayList<HashMap<String, Integer>> countsList = new ArrayList<HashMap<String, Integer>>();
				ArrayList<String> mvTokens = instance.getTokens();
				for (int j=0; j<mvTokens.size(); j++) {
					countsList.add(new HashMap<String, Integer>());
				}
				for (String taggerName : taggersDataInstances.keySet()) {
					InstanceList taggerInstances = taggersDataInstances.get(taggerName).get(filename);
					if (taggerInstances == null || instances.size() != taggerInstances.size()) {
						continue; // skip this tagger for this filename
					}
					TermExtractorInstance taggerInstance = (TermExtractorInstance) taggerInstances.get(i);
					ArrayList<String> taggerLabels = taggerInstance.getLabels();
					if (mvTokens.size() != taggerLabels.size()) {
						continue; // skip this tagger for this instance
					}
					for (int j=0; j<mvTokens.size(); j++) {
						String taggerLabel = taggerLabels.get(j);
						if (!countsList.get(j).containsKey(taggerLabel)) {
							countsList.get(j).put(taggerLabel, 0);
						}
						countsList.get(j).put(taggerLabel, countsList.get(j).get(taggerLabel) + 1);
					}
				}
				
				// majority voting - get most votes labels
				ArrayList<String> mvLabels = new ArrayList<String>();
				String labels = "";
				for (int j=0; j<countsList.size(); j++) {
					HashMap<String, Integer> counts = countsList.get(j);
					int highestCount = -1;
					String highestLabel = null;
					for (String label : counts.keySet()) {
						int count = counts.get(label);
						if (count > highestCount) {
							highestCount = count;
							highestLabel = label;
						}
					}
					mvLabels.add(highestLabel);
					labels += highestLabel + "\n";
					sb.append(instance.getTokens().get(j)+" "+highestLabel+"\n");
				}
				
				Instance mvInstanceTemp = new Instance(sb.toString(), null, instance.getName(), null);
				mvInstances.addThruPipe(mvInstanceTemp);
				mvInstanceTemp = mvInstances.remove(mvInstances.size()-1);
				TermExtractorInstance mvInstance = new TermExtractorInstance(mvInstanceTemp.getData(), mvInstanceTemp.getTarget(), mvInstanceTemp.getName(), mvInstanceTemp.getSource());
				mvInstance.setInstanceId(instance.getInstanceId());
				mvInstance.setTokens(instance.getTokens());
				mvInstance.setLabels(mvLabels);
				mvInstances.add(mvInstance);
			}
			majoratyVotingGroundTruth.put(filename, mvInstances);
		}
		
		PrintWriter pwMV = null;
		try {
			pwMV = new PrintWriter("majority_vote_ground_truth.txt");
		} catch (FileNotFoundException e) {
			e.printStackTrace();
		}
		for (String filename : majoratyVotingGroundTruth.keySet()) {
			InstanceList instances = majoratyVotingGroundTruth.get(filename);
			for (int i=0; i<instances.size(); i++) {
				TermExtractorInstance instance = (TermExtractorInstance) instances.get(i);
				ArrayList<String> tokens = instance.getTokens();
				ArrayList<String> labels = instance.getLabels();
				for (int j=0; j<tokens.size(); j++) {
					pwMV.println(tokens.get(j) + "\t" + labels.get(j));
				}
				pwMV.println();
				pwMV.flush();
			}
		}
		pwMV.flush();
		pwMV.close();
		
		return majoratyVotingGroundTruth;
	}
	
	private TreeMap<String, InstanceList> estimateGroundTruthUsingMajorityVoting1(TreeMap<String, TreeMap<String, InstanceList>> taggersDataInstances, TreeMap<String, InstanceList> distinctInputInstances) {
		System.out.println("[MultipleTaggers] Estimating ground truth using majority voting...");
		TreeMap<String, InstanceList> majoratyVotingGroundTruth = new TreeMap<String, InstanceList>();
		for (String filename : distinctInputInstances.keySet()) {
			InstanceList instances = distinctInputInstances.get(filename);
			
			// iterate over instances and determine majority voting
			InstanceList mvInstances = new InstanceList(sequenceLabeller.getFeaturePipeline()); 
			for (int i=0; i<instances.size(); i++) {
				TermExtractorInstance instance = (TermExtractorInstance) instances.get(i);
				
				// majority voting - get votes
				HashMap<String, Integer> votes = new HashMap<String, Integer>();
				ArrayList<String> mvTokens = instance.getTokens();
				for (String taggerName : taggersDataInstances.keySet()) {
					InstanceList taggerInstances = taggersDataInstances.get(taggerName).get(filename);
					if (taggerInstances == null || instances.size() != taggerInstances.size()) {
						continue; // skip this tagger for this filename
					}
					TermExtractorInstance taggerInstance = (TermExtractorInstance) taggerInstances.get(i);
					if (mvTokens.size() != taggerInstance.getLabels().size()) {
						continue; // skip this tagger for this instance
					}
					String taggerLabelSequence = "";
					for (int j=0; j<mvTokens.size(); j++) {
						String label = taggerInstance.getLabels().get(j);
						taggerLabelSequence += label + "|";
					}
					taggerLabelSequence = taggerLabelSequence.substring(0, taggerLabelSequence.length()-1);
					if (!votes.containsKey(taggerLabelSequence)) {
						votes.put(taggerLabelSequence, 1);
					} else {
						votes.put(taggerLabelSequence, votes.get(taggerLabelSequence) + 1);
					}
				}
				
				// majority voting - get most voted label sequences
				int highestCount = -1;
				String highestLabelSequence = null;
				for (String labelSequence : votes.keySet()) {
					int count = votes.get(labelSequence);
					if (count > highestCount) {
						highestCount = count;
						highestLabelSequence = labelSequence;
					}
				}
				ArrayList<String> mvLabels = new ArrayList<String>();
				String labels = "";
				StringBuffer sb = new StringBuffer();
				String[] splitLabels = highestLabelSequence.split("\\|");
				for (int l=0; l<splitLabels.length; l++) {
					mvLabels.add(splitLabels[l]);
					labels += splitLabels[l] + "\n";
					sb.append(instance.getTokens().get(l)+" "+splitLabels[l]+"\n");
				}
			
				Instance mvInstanceTemp = new Instance(sb.toString(), null, instance.getName(), null);
				mvInstances.addThruPipe(mvInstanceTemp);
				mvInstanceTemp = mvInstances.remove(mvInstances.size()-1);
				TermExtractorInstance mvInstance = new TermExtractorInstance(mvInstanceTemp.getData(), mvInstanceTemp.getTarget(), mvInstanceTemp.getName(), mvInstanceTemp.getSource());
				mvInstance.setInstanceId(instance.getInstanceId());
				mvInstance.setTokens(instance.getTokens());
				mvInstance.setLabels(mvLabels);
				mvInstances.add(mvInstance);
			}
			majoratyVotingGroundTruth.put(filename, mvInstances);
		}
		
		PrintWriter pwMV = null;
		try {
			pwMV = new PrintWriter("majority_vote_ground_truth2.txt");
		} catch (FileNotFoundException e) {
			e.printStackTrace();
		}
		for (String filename : majoratyVotingGroundTruth.keySet()) {
			InstanceList instances = majoratyVotingGroundTruth.get(filename);
			for (int i=0; i<instances.size(); i++) {
				TermExtractorInstance instance = (TermExtractorInstance) instances.get(i);
				ArrayList<String> tokens = instance.getTokens();
				ArrayList<String> labels = instance.getLabels();
				for (int j=0; j<tokens.size(); j++) {
					pwMV.println(tokens.get(j) + "\t" + labels.get(j));
				}
				pwMV.println();
				pwMV.flush();
			}
		}
		pwMV.flush();
		pwMV.close();
		
		return majoratyVotingGroundTruth;
	}
	
	// Evaluation for keyphrase extraction
//	public static EvaluationResults evaluateGroundTruth(TreeMap<String, InstanceList> estimatedGroundTruth, TreeMap<String, InstanceList> actualGroundTruth) {
//		int tokensCount = 0;
//		int correctLabels = 0;
//		TreeMap<String, HashSet<String>> predKeywordsAllDocuments = new TreeMap<String, HashSet<String>>();
//		TreeMap<String, HashSet<String>> gtKeywordsAllDocuments = new TreeMap<String, HashSet<String>>(); 
//		for (String filename : estimatedGroundTruth.keySet()) {
//			InstanceList predInstances = estimatedGroundTruth.get(filename);
//			InstanceList gtInstances = actualGroundTruth.get(filename);
//			HashSet<String> predKeywords = new HashSet<String>();
//			HashSet<String> gtKeywords = new HashSet<String>();
//			if (gtInstances == null || gtInstances.size() != predInstances.size()) {
//				continue;
//			}
//			for (int i=0; i<predInstances.size(); i++) {
//				TermExtractorInstance predInstance = (TermExtractorInstance) predInstances.get(i);
//				TermExtractorInstance gtInstance = (TermExtractorInstance) gtInstances.get(i);
//				ArrayList<String> tokens = predInstance.getTokens();
//				ArrayList<String> predLabels = predInstance.getLabels();
//				//ArrayList<String> gtTokens = gtInstance.getTokens();
//				ArrayList<String> gtLabels = gtInstance.getLabels();
//				if (predLabels.size() != gtLabels.size()) {
//					continue;
//				}
//				for (int j=0; j<tokens.size(); j++) {
//					//System.out.println("DEBUG: " + tokens.get(j) + " GT=" + gtLabels.get(j) + " PRED=" + predLabels.get(j));
//					tokensCount++;
//					if (predLabels.get(j).equals(gtLabels.get(j))) {
//						correctLabels++;
//					}
//				}
//				predKeywords = getKeywords(predKeywords, tokens, predLabels);
//				gtKeywords = getKeywords(gtKeywords, tokens, gtLabels);
//			}
//			predKeywordsAllDocuments.put(filename, predKeywords);
//			gtKeywordsAllDocuments.put(filename, gtKeywords);
//		}
//		double tokenAccuracy = ((double) correctLabels) / ((double) tokensCount); 
//		EvaluationResults evaluationResults = evaluateKeywords(predKeywordsAllDocuments, gtKeywordsAllDocuments);
//		evaluationResults.setTokenAccuracy(tokenAccuracy);
//		evaluationResults.setNumTokens(tokensCount);
//		return evaluationResults;
//	}
	
	// Evaluation for NER
	public static EvaluationResults evaluateAgainstGroundTruth(TreeMap<String, InstanceList> estimatedGroundTruth, TreeMap<String, InstanceList> actualGroundTruth) {
		int numCorrectTokens, totalTokens;
		int[] numTrueSegments, numPredictedSegments, numCorrectSegments;
		int allIndex = START_TAGS.length;
		numTrueSegments = new int[allIndex + 1];
		numPredictedSegments = new int[allIndex + 1];
		numCorrectSegments = new int[allIndex + 1];

		totalTokens = numCorrectTokens = 0;
		for (int n = 0; n < numTrueSegments.length; n++)
			numTrueSegments[n] = numPredictedSegments[n] = numCorrectSegments[n] = 0;
		
		//System.out.println("[MultipleTaggers] estimatedGroundTruth.size()="+estimatedGroundTruth.size());
        //System.out.println("[MultipleTaggers] actualGroundTruth.size()="+actualGroundTruth.size());
		for (String filename : estimatedGroundTruth.keySet()) {
			if (filename.equals(".DS_Store")) {
				continue;
			}
			InstanceList predInstances = estimatedGroundTruth.get(filename);
			InstanceList gtInstances = actualGroundTruth.get(filename);
			//System.out.println("[MultipleTaggers] predInstances.size()="+predInstances.size() + " gtInstances.size()="+gtInstances.size());
			//System.out.println("[MultipleTaggers] filename: " + filename + " estimatedGroundTruthSize:" + estimatedGroundTruth.get(filename).size() + " actualGroundTruthSize:" + actualGroundTruth.get(filename).size());
			
			if (predInstances.size() != gtInstances.size()) {
				//throw new RuntimeException();
				System.out.println("[MultipleTaggers] SKIPPING FILENAME:" + filename);
				continue;
			}
			
			int trueStart, predStart; // -1 for non-start, otherwise index into segmentStartTag
			for (int i=0; i<predInstances.size(); i++) {
				TermExtractorInstance predInstance = (TermExtractorInstance) predInstances.get(i);
				TermExtractorInstance gtInstance = (TermExtractorInstance) gtInstances.get(i);
				ArrayList<String> tokens = predInstance.getTokens();
				ArrayList<String> predLabels = predInstance.getLabels();
				ArrayList<String> gtLabels = gtInstance.getLabels();
				if (predLabels.size() != gtLabels.size()) {
					//System.out.println("[MultipleTaggers]- \tSKIP");
					continue; // ok, ignore it... solutions??
				}
				
				for (int j=0; j<tokens.size(); j++) {
					//System.out.println("[MultipleTaggers]- \tT="+tokens.get(j)+" GT="+gtLabels.get(j)+" P="+predLabels.get(j));
					totalTokens++;
					if (gtLabels.get(j).equals(predLabels.get(j))) {
						numCorrectTokens++;
					}
					trueStart = predStart = -1;
					// Count true segment starts
					for (int n = 0; n < START_TAGS.length; n++) {
						if (START_TAGS[n].equals(gtLabels.get(j))) {
							numTrueSegments[n]++;
							numTrueSegments[allIndex]++;
							trueStart = n;
							break;
						}
					}
					// Count predicted segment starts
					for (int n = 0; n < START_TAGS.length; n++) {
						if (START_TAGS[n].equals(predLabels.get(j))) {
							numPredictedSegments[n]++;
							numPredictedSegments[allIndex]++;
							predStart = n;
						}
					}
					if (trueStart != -1 && trueStart == predStart) {
						// Truth and Prediction both agree that the same segment
						// tag-type is starting now
						int m;
						boolean trueContinue = false;
						boolean predContinue = false;
						for (m = j + 1; m < gtLabels.size(); m++) {
							trueContinue = CONTINUE_TAGS[predStart].equals(gtLabels.get(m));
							predContinue = CONTINUE_TAGS[predStart].equals(predLabels.get(m));
							if (!trueContinue || !predContinue) {
								if (trueContinue == predContinue) {
									// They agree about a segment is ending somehow
									numCorrectSegments[predStart]++;
									numCorrectSegments[allIndex]++;
								}
								break;
							}
						}
						// for the case of the end of the sequence
						if (m == gtLabels.size()) {
							if (trueContinue == predContinue) {
								numCorrectSegments[predStart]++;
								numCorrectSegments[allIndex]++;
							}
						}
					}
				}
			}
		}
		
		// Compute precision, recall and f1
		double tokenAccuracy = ((double) numCorrectTokens) / ((double) totalTokens);
		double precision = numPredictedSegments[allIndex] == 0 ? 1
				: ((double) numCorrectSegments[allIndex]) / numPredictedSegments[allIndex];
		double recall = numTrueSegments[allIndex] == 0 ? 1
				: ((double) numCorrectSegments[allIndex]) / numTrueSegments[allIndex];
		double f1 = recall + precision == 0.0 ? 0.0
				: (2.0 * recall * precision) / (recall + precision);
		
		// Return evalutation results
		EvaluationResults evaluationResults = new EvaluationResults(precision, recall, f1);
		evaluationResults.setTokenAccuracy(tokenAccuracy);
		evaluationResults.setNumTokens(totalTokens);
		return evaluationResults;
	}
	
	public static HashSet<String> getKeywords(HashSet<String> keywords, ArrayList<String> tokens, ArrayList<String> labels) {
		assert (tokens.size() == labels.size());
		String kw = null;
		for (int i=0; i<tokens.size(); i++) {
			String token = tokens.get(i);
			String label = labels.get(i);
			
			if (label.equals("B-KEYWORD")) {
				if (kw != null) {
					keywords.add(kw);
				}
				kw = token;
			} else if (label.equals("I-KEYWORD")) {
				kw += " " + token;
			} else {
				if (kw != null) {
					keywords.add(kw);
					kw = null;
				}
			}
		}
		if (kw != null) {
			keywords.add(kw);
		}
		return keywords;
	}
	
	public static EvaluationResults evaluateKeywords(TreeMap<String, HashSet<String>> predKeywordsAllDocuments, TreeMap<String, HashSet<String>> gtKeywordsAllDocuments) {
	    int numTrueSegments = 0;
	    int numPredictedSegments = 0; 
	    int numCorrectSegments = 0;
		assert (predKeywordsAllDocuments.size() == gtKeywordsAllDocuments.size());
		for (String filename : predKeywordsAllDocuments.keySet()) {
			HashSet<String> predKeywords = predKeywordsAllDocuments.get(filename);
			HashSet<String> gtKeywords = gtKeywordsAllDocuments.get(filename);
			assert (predKeywords.size() == gtKeywords.size());
			
	    	numPredictedSegments += predKeywords.size();
	    	numTrueSegments += gtKeywords.size();
	    	
	    	for (String kw : predKeywords) {
	    		if (gtKeywords.contains(kw)) {
	    			numCorrectSegments++;
	    		}
	    	}
		}
	    
		double precision = numPredictedSegments == 0 ? 1 : ((double)numCorrectSegments) / numPredictedSegments;
		double recall = numTrueSegments == 0 ? 1 : ((double)numCorrectSegments) / numTrueSegments;
		double f1 = recall+precision == 0.0 ? 0.0 : (2.0 * recall * precision) / (recall + precision);
		return new EvaluationResults(precision, recall, f1);
	}
	
	public void normalTrain() {
		System.out.flush();
		TreeMap<String, Double> trainingPercentages = new TreeMap<String, Double>(); // training percentages of data from each tagger to use for training in each iteration
		for (String taggerName : taggersDataInstances.keySet()) {
			trainingPercentages.put(taggerName, 1.0);
		}
		
		// get instances for training 
		TreeMap<String, TreeMap<String, InstanceList>> trainingData = new TreeMap<String, TreeMap<String, InstanceList>>();
		for (String taggerName : taggersDataInstances.keySet()) {
			trainingData.put(taggerName, taggersDataInstances.get(taggerName));
		}
		
		// train CRF
		EvaluationResults[] evaluationResults = trainCRF(trainingData, trainingPercentages);
		EvaluationResults gtEvaluationResults = evaluationResults[0];
		EvaluationResults testDataEvaluationResults = evaluationResults[1];

		DecimalFormat f = new DecimalFormat("0.####");
		if (logPW != null) {
			logPW.println("NORMAL CRF (train): " + f.format(gtEvaluationResults.getPrecision()) + "\t" + f.format(gtEvaluationResults.getRecall()) + "\t" + f.format(gtEvaluationResults.getF1()));
			if (testDataEvaluationResults != null) {
				logPW.println("NORMAL CRF (test): " + f.format(testDataEvaluationResults.getPrecision()) + "\t" + f.format(testDataEvaluationResults.getRecall()) + "\t" + f.format(testDataEvaluationResults.getF1()));
			}
			logPW.flush();
		}
	}
	
	public void groundTruthTraining() {
		System.out.flush();
		TreeMap<String, Double> trainingPercentages = new TreeMap<String, Double>(); // training percentages of data from each tagger to use for training in each iteration
		trainingPercentages.put("GOLD", 1.0);
		
		// get instances for training 
		TreeMap<String, TreeMap<String, InstanceList>> trainingData = new TreeMap<String, TreeMap<String, InstanceList>>();
		trainingData.put("GOLD", actualGroundTruth);
		
		// train CRF
		EvaluationResults[] evaluationResults = trainCRF(trainingData, trainingPercentages);
		EvaluationResults gtEvaluationResults = evaluationResults[0];
		EvaluationResults testDataEvaluationResults = evaluationResults[1];

		DecimalFormat f = new DecimalFormat("0.####");
		if (logPW != null) {
			logPW.println("GOLD CRF (train): " + f.format(gtEvaluationResults.getPrecision()) + "\t" + f.format(gtEvaluationResults.getRecall()) + "\t" + f.format(gtEvaluationResults.getF1()));
			if (testDataEvaluationResults != null) {
				logPW.println("GOLD CRF (test): " + f.format(testDataEvaluationResults.getPrecision()) + "\t" + f.format(testDataEvaluationResults.getRecall()) + "\t" + f.format(testDataEvaluationResults.getF1()));
			}
			logPW.flush();
		}
	}
		
	public EvaluationResults[] trainCRF(TreeMap<String, TreeMap<String, InstanceList>> trainingData, TreeMap<String, Double> trainingPercentages) {
		initializeTermExtractor();
		
		// train model
		System.out.print("[MultipleTaggers]\n[MultipleTaggers] Training normal CRF model... ");
		System.out.flush();
		sequenceLabeller.setIterations(Integer.MAX_VALUE); // variable number of iterations
		sequenceLabeller.setGaussianVariance(10.0); // fixed variance
		try {
			InstanceList trainingDataInstances = sequenceLabeller.trainModel(trainingData, nthreads, Integer.MAX_VALUE, trainingPercentages);
			double trainingDataLikelihood = sequenceLabeller.loglikelihood(trainingDataInstances);
			System.out.println("[MultipleTaggers] \tTraining instances likelihood: " + trainingDataLikelihood);
		} catch (IOException e) {
			e.printStackTrace();
		}
		
		// apply model
		PrintWriter pw = null;
		try {
			pw = new PrintWriter("estimatedGroundTruthCRF.txt");
		} catch (FileNotFoundException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}
		TreeMap<String, InstanceList> estimatedGroundTruth = sequenceLabeller.applyModelToInstances(distinctInputInstances, pw, false);
		
		// evaluate the current ground truth estimate
		System.out.println("[MultipleTaggers] \tMeasuring estimated ground truth accuracy... ");
		DecimalFormat f = new DecimalFormat("0.####");
		EvaluationResults gtEvaluationResults = CRFWithMultipleTaggers.evaluateAgainstGroundTruth(estimatedGroundTruth, actualGroundTruth);
		System.out.println("[MultipleTaggers] \tResults: tokenAcc="+f.format(gtEvaluationResults.getTokenAccuracy())+
				" precision="+f.format(gtEvaluationResults.getPrecision())+
				" recall="+f.format(gtEvaluationResults.getRecall())+" f1="+f.format(gtEvaluationResults.getF1()));
		System.out.flush();
		
		// evaluate model against test data
		EvaluationResults testDataEvaluationResults = null;
		if (evaluationData != null) { 
			System.out.println("[MultipleTaggers] Evaluating model on test data... ");
			testDataEvaluationResults = testModel(evaluationData);
			System.out.println("[MultipleTaggers] Results: tokenAcc="+f.format(testDataEvaluationResults.getTokenAccuracy())+
					" precision="+f.format(testDataEvaluationResults.getPrecision())+
					" recall="+f.format(testDataEvaluationResults.getRecall())+" f1="+f.format(testDataEvaluationResults.getF1()));
		}
		
		return new EvaluationResults[]{gtEvaluationResults, testDataEvaluationResults};
	}
	
	public EvaluationResults testModel(TreeMap<String, InstanceList> testData) {
		TreeMap<String, InstanceList> testDataPred = sequenceLabeller.applyModelToInstances(testData, null, false);
		EvaluationResults evaluationResults = CRFWithMultipleTaggers.evaluateAgainstGroundTruth(testData, testDataPred);
		return evaluationResults;
	}
	
	public void mvTrain() {
		mtTrainer = new MTTrainerMajorityVoting(inputFolder, sequenceLabeller, nthreads, taggersDataInstances, totalDataInstances, distinctInputInstances, actualGroundTruth, evaluationData);
		double updateRate = 1.0;
		mtTrainer.train(maxIterations);
	}
	
	public void emTrain() {
		initializeTermExtractor();
		mtTrainer = new MTTrainerF1(inputFolder, sequenceLabeller, nthreads, taggersDataInstances, totalDataInstances, distinctInputInstances, actualGroundTruth, evaluationData, majoratyVotingGroundTruth2);
		//mtTrainer = new MTTrainerF1Dirichlet(inputFolder, sequenceLabeller, nthreads, taggersDataInstances, totalDataInstances, distinctInputInstances, actualGroundTruth, evaluationData, majoratyVotingGroundTruth2);
		//mtTrainer = new MTTrainerMajorityVoting(inputFolder, sequenceLabeller, nthreads, taggersDataInstances, totalDataInstances, distinctInputInstances, actualGroundTruth, evaluationData);
		mtTrainer.train(maxIterations);
	}

	public static void main(String[] args) throws FileNotFoundException {
		// Disable evil mallet logger
		String propThatKeepsMalletFromMessingWithLogging = "java.util.logging.config.file";
		//String propValue = System.getProperty(propThatKeepsMalletFromMessingWithLogging);
		System.setProperty(propThatKeepsMalletFromMessingWithLogging, "anything-but-null");
		System.out.flush();
		
		//String inputFolder = "mturk_ner_data/extracted_data";
		//String groundTruthFolder = "mturk_ner_data/ground_truth";
		//String evaluationFolder = "mturk_ner_data/test_data";
		String inputFolder = "data/mturk/extracted_data"; 
		//String inputFolder = "data/mturk/subsample";
		String groundTruthFolder = "data/mturk/ground_truth";
		//String evaluationFolder = null;
		String evaluationFolder = "data/mturk/test";
		int maxIterations = 8; 
		int nthreads = 4;
		CRFWithMultipleTaggers.logPW = new PrintWriter("single_results.txt");
		CRFWithMultipleTaggers.logPW.println("\nInput file: " + inputFolder);
		CRFWithMultipleTaggers.logPW.flush();
		
		CRFWithMultipleTaggers crfma = new CRFWithMultipleTaggers(inputFolder, groundTruthFolder, evaluationFolder, maxIterations, nthreads);
		crfma.groundTruthTraining();
		crfma.normalTrain();
		logPW.flush();
		CRFWithMultipleTaggers.logPW.println("Starting CRF-EM...");
		CRFWithMultipleTaggers.logPW.flush();
		crfma.emTrain(); 
		CRFWithMultipleTaggers.logPW.flush();
		CRFWithMultipleTaggers.logPW.close();
	}
}
