package pt.uc.dei.crf_ma;

import java.io.FileNotFoundException;
import java.io.IOException;
import java.io.PrintWriter;
import java.text.DecimalFormat;
import java.util.TreeMap;


import cc.mallet.types.InstanceList;

public class MTTrainerF1 extends MTTrainer {
	
	private TreeMap<String, InstanceList> majorityVotingGroundTruth;
	
	public MTTrainerF1(
			String inputFolder,
			SequenceLabeller sequenceLabeller,
			int nthreads,
			TreeMap<String, TreeMap<String, InstanceList>> taggersDataInstances,
			int totalDataInstances,
			TreeMap<String, InstanceList> distinctInputInstances,
			TreeMap<String, InstanceList> actualGroundTruth,
			TreeMap<String, InstanceList> evaluationData,
			TreeMap<String, InstanceList> majorityVotingGroundTruth) {
		super(inputFolder, sequenceLabeller, nthreads, taggersDataInstances,
				totalDataInstances, distinctInputInstances, actualGroundTruth, evaluationData);
		this.majorityVotingGroundTruth = majorityVotingGroundTruth;
	}
	
	public TreeMap<String, Double> majorityVotingInitialization() {
		// evaluate taggers accuracies
		System.out.println("[MultipleTaggers] Majority Voting initialization...");
		TreeMap<String, Double> f1Scores = new TreeMap<String, Double>();
		double sum_f1 = 0;
		double sum_f1_squared = 0;
		for (String taggerName : taggersDataInstances.keySet()) {
			TreeMap<String, InstanceList> taggerInstances = taggersDataInstances.get(taggerName);
			EvaluationResults taggerEvaluationResults = evaluateTagger(taggerName, taggerInstances, majorityVotingGroundTruth);
			sum_f1 += taggerEvaluationResults.getF1();
			sum_f1_squared += (taggerEvaluationResults.getF1() * taggerEvaluationResults.getF1());
			f1Scores.put(taggerName,  taggerEvaluationResults.getF1());
		}
		TreeMap<String, Double> trainingPercentages = new TreeMap<String, Double>(); // training percentages of data from each tagger to use for training in each iteration
		for (String taggerName : taggersDataInstances.keySet()) {
			double f1 = f1Scores.get(taggerName);
			double pi = f1 * f1 / sum_f1_squared;
			trainingPercentages.put(taggerName, pi);
		}
		return trainingPercentages;
	}

	@Override
	public TreeMap<String, InstanceList> train(int maxIterations) {
		System.out.println("[MultipleTaggers] MTTrainerF1 starting...");
		System.out.flush();
		DecimalFormat f = new DecimalFormat("0.####");
		TreeMap<String, Double> trainingPercentages = new TreeMap<String, Double>(); // training percentages of data from each tagger to use for training in each iteration
		TreeMap<String, Double> pis = new TreeMap<String, Double>(); 
		TreeMap<String, Double> gamma_numerators = new TreeMap<String, Double>(); 
		TreeMap<String, Double> gammas = new TreeMap<String, Double>(); 
		
		// initialization...
		trainingPercentages = majorityVotingInitialization();
		for (String taggerName : taggersDataInstances.keySet()) {
			//trainingPercentages.put(taggerName, 1.0);
			//trainingPercentages.put(taggerName, 1.0 / taggerInstances.size());
			gamma_numerators.put(taggerName, trainingPercentages.get(taggerName));
			gammas.put(taggerName, trainingPercentages.get(taggerName));
		}
		
		for(int iteration = 0; iteration < maxIterations; iteration++) {
			System.out.println("[MultipleTaggers] \n[MultipleTaggers] Iteration: " + iteration);
			System.out.flush();
			String gtResultsLine = iteration + " (train) ";
			
			// get instances for training 
			TreeMap<String, TreeMap<String, InstanceList>> trainingData = new TreeMap<String, TreeMap<String, InstanceList>>();
			for (String taggerName : taggersDataInstances.keySet()) {
				trainingData.put(taggerName, taggersDataInstances.get(taggerName));
			}

			// train model
			System.out.print("[MultipleTaggers] Training CRF model... ");
			System.out.flush();
			sequenceLabeller.setIterations(Integer.MAX_VALUE); // variable number of iterations
			sequenceLabeller.setGaussianVariance(10.0); // fixed variance
			try {
				InstanceList trainingDataInstances = sequenceLabeller.trainModel(trainingData, nthreads, Integer.MAX_VALUE, trainingPercentages);
				System.out.println(" (training data size=" + trainingDataInstances.size() + ")");
				double trainingDataLikelihood = sequenceLabeller.loglikelihood(trainingDataInstances);
				System.out.println("[MultipleTaggers] CRF training converged after " + sequenceLabeller.usedIterations + " iterations"); 
				System.out.println("[MultipleTaggers] Training instances likelihood: " + trainingDataLikelihood);
			} catch (IOException e) {
				e.printStackTrace();
			}

			if (sequenceLabeller.usedIterations < 5) {
				for (String taggerName : taggersDataInstances.keySet()) {
                        		trainingPercentages.put(taggerName, trainingPercentages.get(taggerName) + 0.01);
                        		//trainingPercentages.put(taggerName, 1.0 / taggerInstances.size());
                        		gamma_numerators.put(taggerName, trainingPercentages.get(taggerName));
                        		gammas.put(taggerName, trainingPercentages.get(taggerName));
                		}
				System.out.println("[MultipleTaggers] RE-RUN CRF training!");
				continue;
			}
			
			// estimate new ground truth
			estimatedGroundTruth = estimateGroundTruth(distinctInputInstances);
			
			// evaluate the current ground truth estimate
			System.out.println("[MultipleTaggers] Measuring estimated ground truth accuracy... ");
			EvaluationResults gtEvaluationResults = CRFWithMultipleTaggers.evaluateAgainstGroundTruth(estimatedGroundTruth, actualGroundTruth);
			gtResultsLine += "\t"+f.format(gtEvaluationResults.getPrecision())+"\t"+
					f.format(gtEvaluationResults.getRecall())+"\t"+f.format(gtEvaluationResults.getF1());
			System.out.println("[MultipleTaggers] \tResults: tokenAcc="+f.format(gtEvaluationResults.getTokenAccuracy())+
					" precision="+f.format(gtEvaluationResults.getPrecision())+
					" recall="+f.format(gtEvaluationResults.getRecall())+" f1="+f.format(gtEvaluationResults.getF1()));
			System.out.flush();
			
			// evaluate model against test data
			String testDataResultsLine = null;
			if (evaluationData != null) { 
				System.out.println("[MultipleTaggers] Evaluating model on test data... ");
				EvaluationResults testDataEvaluationResults = testModel(evaluationData);
				System.out.println("[MultipleTaggers] \tResults: tokenAcc="+f.format(testDataEvaluationResults.getTokenAccuracy())+
						" precision="+f.format(testDataEvaluationResults.getPrecision())+
						" recall="+f.format(testDataEvaluationResults.getRecall())+" f1="+f.format(testDataEvaluationResults.getF1()));
				testDataResultsLine = iteration + " (test) \t"+f.format(testDataEvaluationResults.getPrecision())+"\t"+
						f.format(testDataEvaluationResults.getRecall())+"\t"+f.format(testDataEvaluationResults.getF1());
			}

			// compute log likelihoods
			System.out.println("[MultipleTaggers] Computing loglikelihoods...");
			System.out.flush();
			TreeMap<String, Double> loglikelihoods = new TreeMap<String, Double>();
			for (String taggerName : taggersDataInstances.keySet()) {
				try {
					double loglikelihood = sequenceLabeller.loglikelihood(inputFolder + "" + taggerName);
					loglikelihoods.put(taggerName, loglikelihood);
					//System.out.println("[MultipleTaggers] Likelihood from tagger " + taggerName + " ["+i+"/"+taggersDataInstances.keySet().size()+"]: " + loglikelihood);
				} catch (IOException e) {
					e.printStackTrace();
				}
			}
			
			// determine basic stats of the loglikelihoods 
			double min_ll = Integer.MAX_VALUE;
			double max_ll = Integer.MIN_VALUE;
			double sum_ll = 0;
			for (String taggerName : loglikelihoods.keySet()) {
				double ll = loglikelihoods.get(taggerName);
				sum_ll += ll;
				if (ll > max_ll) {
					max_ll = ll;
				}
				if (ll < min_ll) {
					min_ll = ll;
				}
			}
			
			// convert loglikelihoods to the [0,1] interval
			TreeMap<String, Double> normalizedLoglikelihoods = new TreeMap<String, Double>();
			double sum_norm_ll = 0;
			for (String taggerName : loglikelihoods.keySet()) {
				double ll = loglikelihoods.get(taggerName);
				double norm_ll = (ll - min_ll) / (max_ll - min_ll);
				sum_norm_ll += norm_ll;
				normalizedLoglikelihoods.put(taggerName, norm_ll);
			}
			
			// evaluate taggers accuracies -> compute f1's
			System.out.println("[MultipleTaggers] Evaluating taggers...");
			TreeMap<String, Double> f1Scores = new TreeMap<String, Double>();
			double sum_f1 = 0;
			double sum_f1_squared = 0;
			double modelLikelihood = 0;
			for (String taggerName : taggersDataInstances.keySet()) {
				TreeMap<String, InstanceList> taggerInstances = taggersDataInstances.get(taggerName);
				EvaluationResults taggerEvaluationResults = evaluateTagger(taggerName, taggerInstances, estimatedGroundTruth);
				double f1 = taggerEvaluationResults.getF1();
				f1Scores.put(taggerName,  f1);
				sum_f1 += f1;
				sum_f1_squared += (f1 * f1);
				modelLikelihood += (taggerEvaluationResults.getF1() * loglikelihoods.get(taggerName)) * -1;
			}
			
			// compute pi's numerators and denominator
			double gamma_denominator = 0;
			for (String taggerName : taggersDataInstances.keySet()) {
				double f1 = f1Scores.get(taggerName);
				double pi = (f1 * f1) / sum_f1_squared;
				pis.put(taggerName, pi);
				double gamma_numerator = pi * normalizedLoglikelihoods.get(taggerName);
				gamma_numerators.put(taggerName, gamma_numerator);
				gamma_denominator += gamma_numerator;
			}
			
			// compute gamma's
			for (String taggerName : taggersDataInstances.keySet()) {
				//double gamma = gamma_numerators.get(taggerName) / gamma_denominator;
				double gamma = gamma_numerators.get(taggerName);
				gammas.put(taggerName, gamma);
			}
			
			// update training percentages
			System.out.println("[MultipleTaggers]- Training percentages updates:");
			double modelLikelihood2 = 0;
			double modelLikelihood3 = 0;
			double modelLikelihood4 = 0;
			for (String taggerName : taggersDataInstances.keySet()) {
				double ll = loglikelihoods.get(taggerName);
				double norm_ll = normalizedLoglikelihoods.get(taggerName);
				double f1 = f1Scores.get(taggerName);
				double pi = pis.get(taggerName);
				double gamma = gammas.get(taggerName);

				double oldPercentage = trainingPercentages.get(taggerName);
				double newPercentage = gamma;
				System.out.println("[MultipleTaggers]- 	"+taggerName+": f1="+f.format(f1)+" pi="+f.format(pi)+" norm_ll="+f.format(norm_ll)+" gamma_nom="+f.format(gamma*gamma_denominator)+
						" gamma="+f.format(gamma));
				trainingPercentages.put(taggerName, newPercentage); // new version (weighted training instances)
				modelLikelihood2 += (gamma/gamma_denominator) * ll * -1;
				modelLikelihood3 += (f1*f1/sum_f1_squared) * ll * -1;
				modelLikelihood4 += gamma;
			}
			System.out.println("[MultipleTaggers] Model likelihood: " + modelLikelihood);
			System.out.println("[MultipleTaggers] Model likelihood2: " + modelLikelihood2);
			System.out.println("[MultipleTaggers] Model likelihood3: " + modelLikelihood3);
			System.out.println("[MultipleTaggers] Model likelihood4: " + modelLikelihood4);
			System.out.flush();
			gtResultsLine += "\t"+f.format(modelLikelihood)+"\t"+f.format(modelLikelihood2)+"\t"+f.format(modelLikelihood3)+"\t"+f.format(modelLikelihood4);
			pwStats.println(gtResultsLine);
			pwStats.flush();
			if (CRFWithMultipleTaggers.logPW != null) {
				CRFWithMultipleTaggers.logPW.println(gtResultsLine);
				if (testDataResultsLine != null) {
					CRFWithMultipleTaggers.logPW.println(testDataResultsLine);
				}
				CRFWithMultipleTaggers.logPW.flush();
			}
		}
		pwStats.close();
		
		return estimatedGroundTruth;
	}
	
	public TreeMap<String, InstanceList> estimateGroundTruth(TreeMap<String, InstanceList> distinctInputInstances) {
		System.out.println("[MultipleTaggers] Estimating new ground truth...");
		PrintWriter pwGT = null;
		try {
			pwGT = new PrintWriter("estimated_ground_truth.txt");
		} catch (FileNotFoundException e) {
			e.printStackTrace();
		}
		TreeMap<String, InstanceList> estimatedGroundTruth = sequenceLabeller.applyModelToInstances(distinctInputInstances, pwGT, false);
		return estimatedGroundTruth;
	}
	
	public EvaluationResults testModel(TreeMap<String, InstanceList> testData) {
		TreeMap<String, InstanceList> testDataPred = sequenceLabeller.applyModelToInstances(testData, null, false);
		EvaluationResults evaluationResults = CRFWithMultipleTaggers.evaluateAgainstGroundTruth(testData, testDataPred);
		return evaluationResults;
	}
}
