package pt.uc.dei.crf_ma;

import java.io.FileNotFoundException;
import java.io.PrintWriter;
import java.text.DecimalFormat;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.TreeMap;

import pt.uc.dei.crf_ma.types.TermExtractorInstance;

import cc.mallet.types.InstanceList;

public class MTTrainerMajorityVoting extends MTTrainer {
	
	public MTTrainerMajorityVoting(
			String inputFolder,
			SequenceLabeller sequenceLabeller,
			int nthreads,
			TreeMap<String, TreeMap<String, InstanceList>> taggersDataInstances,
			int totalDataInstances,
			TreeMap<String, InstanceList> distinctInputInstances,
			TreeMap<String, InstanceList> actualGroundTruth,
			TreeMap<String, InstanceList> evaluationData) {
		super(inputFolder, sequenceLabeller, nthreads, taggersDataInstances,
				totalDataInstances, distinctInputInstances, actualGroundTruth, evaluationData);
	}

	@Override
	public TreeMap<String, InstanceList> train(int maxIterations) {
		DecimalFormat f = new DecimalFormat("0.####");
		TreeMap<String, Double> annotatorsQualities = new TreeMap<String, Double>(); // training percentages of data from each tagger to use for training in each iteration
		// initialization...
		for (String taggerName : taggersDataInstances.keySet()) {
			annotatorsQualities.put(taggerName, 1.0);
		}
		
		for(int iteration = 0; iteration<maxIterations; iteration++) {
			System.out.println("[MultipleTaggers]\n[MultipleTaggers] Iteration: " + iteration);
			pwStats.print(iteration);
			
			// use weighted majority voting to estimate ground truth
			System.out.print("[MultipleTaggers] Using Weighted Majority Voting to estimate current ground truth... ");
			TreeMap<String, InstanceList> estimatedGroundTruth = weightedMajorityVoting(taggersDataInstances, distinctInputInstances, annotatorsQualities);
			
			// evaluate the current ground truth estimate
			System.out.println("[MultipleTaggers] Measuring estimated ground truth accuracy... ");
			EvaluationResults gtEvaluationResults = CRFWithMultipleTaggers.evaluateAgainstGroundTruth(estimatedGroundTruth, actualGroundTruth);
			pwStats.print("\t"+f.format(gtEvaluationResults.getTokenAccuracy())+"\t"+f.format(gtEvaluationResults.getPrecision())+"\t"+
					f.format(gtEvaluationResults.getRecall())+"\t"+f.format(gtEvaluationResults.getF1()));
			System.out.println("[MultipleTaggers] Results: tokenAcc="+f.format(gtEvaluationResults.getTokenAccuracy())+
					" precision="+f.format(gtEvaluationResults.getPrecision())+
					" recall="+f.format(gtEvaluationResults.getRecall())+" f1="+f.format(gtEvaluationResults.getF1()));

			if (CRFWithMultipleTaggers.logPW != null) {
				CRFWithMultipleTaggers.logPW.println(iteration + " (train) \t" + f.format(gtEvaluationResults.getPrecision()) + "\t" + f.format(gtEvaluationResults.getRecall()) + "\t" + f.format(gtEvaluationResults.getF1()));
				CRFWithMultipleTaggers.logPW.flush();
			}

			// evaluate annotators accuracies
			System.out.println("[MultipleTaggers] Evaluating taggers...");
			for (String taggerName : taggersDataInstances.keySet()) {
				TreeMap<String, InstanceList> taggerInstances = taggersDataInstances.get(taggerName);
				EvaluationResults taggerEvaluationResults = evaluateTagger(taggerName, taggerInstances, estimatedGroundTruth);
				annotatorsQualities.put(taggerName, taggerEvaluationResults.getF1());
			}
			
			pwStats.flush();
		}
		pwStats.flush();
		pwStats.close();
		
		return estimatedGroundTruth;
	}
	
	private TreeMap<String, InstanceList> weightedMajorityVoting(TreeMap<String, TreeMap<String, InstanceList>> taggersDataInstances, TreeMap<String, InstanceList> distinctInputInstances, TreeMap<String, Double> annotatorsQualities) {
		System.out.println("[MultipleTaggers] Estimating ground truth using majority voting...");
		TreeMap<String, InstanceList> majoratyVotingGroundTruth = new TreeMap<String, InstanceList>();
		for (String filename : distinctInputInstances.keySet()) {
			InstanceList instances = distinctInputInstances.get(filename);
			
			// iterate over instances and determine majority voting
			InstanceList mvInstances = instances.cloneEmpty(); 
			for (int i=0; i<instances.size(); i++) {
				TermExtractorInstance instance = (TermExtractorInstance) instances.get(i);
				
				// make a deep copy of the instance
				TermExtractorInstance mvInstance = new TermExtractorInstance(instance.getData(), instance.getTarget(), instance.getName(), instance.getSource());
				mvInstance.setInstanceId(instance.getInstanceId());
				mvInstance.setTokens(instance.getTokens());
				mvInstance.setLabels(null);
				
				// majority voting - get votes
				ArrayList<HashMap<String, Double>> countsList = new ArrayList<HashMap<String, Double>>();
				ArrayList<String> mvTokens = mvInstance.getTokens();
				for (int j=0; j<mvTokens.size(); j++) {
					countsList.add(new HashMap<String, Double>());
				}
				for (String taggerName : taggersDataInstances.keySet()) {
					InstanceList taggerInstances = taggersDataInstances.get(taggerName).get(filename);
					if (taggerInstances == null || instances.size() != taggerInstances.size()) {
						continue; // skip this tagger for this filename
					}
					TermExtractorInstance taggerInstance = (TermExtractorInstance) taggerInstances.get(i);
					ArrayList<String> taggerLabels = taggerInstance.getLabels();
					if (mvTokens.size() != taggerLabels.size()) {
						continue; // skip this tagger for this instance
					}
					for (int j=0; j<mvTokens.size(); j++) {
						String taggerLabel = taggerLabels.get(j);
						if (!countsList.get(j).containsKey(taggerLabel)) {
							countsList.get(j).put(taggerLabel, 0.0);
						}
						countsList.get(j).put(taggerLabel, countsList.get(j).get(taggerLabel) + annotatorsQualities.get(taggerName));
					}
				}
				
				// majority voting - get most votes labels
				ArrayList<String> mvLabels = new ArrayList<String>();
				for (int j=0; j<countsList.size(); j++) {
					HashMap<String, Double> counts = countsList.get(j);
					double highestCount = -1;
					String highestLabel = null;
					for (String label : counts.keySet()) {
						double count = counts.get(label);
						if (count > highestCount) {
							highestCount = count;
							highestLabel = label;
						}
					}
					mvLabels.add(highestLabel);
				}
				
				mvInstance.setLabels(mvLabels);
				mvInstances.add(mvInstance);
			}
			majoratyVotingGroundTruth.put(filename, mvInstances);
		}
		
		PrintWriter pwMV = null;
		try {
			pwMV = new PrintWriter("majority_vote_ground_truth.txt");
		} catch (FileNotFoundException e) {
			e.printStackTrace();
		}
		for (String filename : majoratyVotingGroundTruth.keySet()) {
			InstanceList instances = majoratyVotingGroundTruth.get(filename);
			for (int i=0; i<instances.size(); i++) {
				TermExtractorInstance instance = (TermExtractorInstance) instances.get(i);
				ArrayList<String> tokens = instance.getTokens();
				ArrayList<String> labels = instance.getLabels();
				for (int j=0; j<tokens.size(); j++) {
					pwMV.println(tokens.get(j) + "\t" + labels.get(j));
				}
				pwMV.println();
				pwMV.flush();
			}
		}
		pwMV.flush();
		pwMV.close();
		
		return majoratyVotingGroundTruth;
	}
}
