package pt.uc.dei.crf_ma.clustering;

import java.io.BufferedReader;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.FileReader;
import java.io.IOException;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.io.PrintWriter;
import java.io.Serializable;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.HashMap;
import java.util.Set;

import pt.uc.dei.crf_ma.SequenceLabeller;

public class CRFHierarchicalClustering implements Serializable {
	
	private static final long serialVersionUID = 7526472295622776147L;
	
	public static final int LINKAGE_SINGLE = 0;
	public static final int LINKAGE_COMPLETE = 1;
	public static final int LINKAGE_UPGMA = 2;
	public static final int LINKAGE_WPGMA = 3;
	public static final int LINKAGE_WARD = 4;

	private HashMap<String, String> dataFiles;
	private int numInstances;
	private int numThreads = 4;
	private HashMap<String, Integer> indexMap;
	private HashMap<Integer, String> labelMap;
	private double[][] logScoreMatrix;
	private double[][] logZMatrix;
	private double[][] probMatrix;
	private double[][] loglikelihoodMatrix;
	private HNode root;
	
	public CRFHierarchicalClustering(String inputFolder, int numThreads) {
		this.dataFiles = getDataFiles(inputFolder);
		this.numThreads = numThreads;
		this.numInstances = dataFiles.size();
	}
	
	private HashMap<String, String> getDataFiles(String inputFolder) {
		HashMap<String, String> dataFiles = new HashMap<String, String>();
		
		File dir = new File(inputFolder);
		if (!inputFolder.endsWith("/")) {
			inputFolder += "/";
		}
		String[] children = dir.list();
		if (children == null) {
		    throw new RuntimeException("Data directory not found: " + inputFolder);
		} else {
		    for (int i=0; i<children.length; i++) {
		        // Get filename of file or directory
		        String filename = children[i];
		        if (!filename.equals(".DS_Store") && !filename.equals("._.DS_Store")) {
		        	dataFiles.put(filename, new String(inputFolder + filename));
		        }
		    }
		}
		
		return dataFiles;
	}
	
	public double[][] calcLogLikelihoods() {
		System.out.println("[CLUSTERING] Building loglikelihoods matrix...");
		this.logScoreMatrix = new double[this.numInstances][this.numInstances];
		this.logZMatrix = new double[this.numInstances][this.numInstances];
		this.probMatrix = new double[this.numInstances][this.numInstances];
		this.loglikelihoodMatrix = new double[this.numInstances][this.numInstances];
		this.indexMap = new HashMap<String, Integer>();
		this.labelMap = new HashMap<Integer, String>();
		int i = 0;
		for (String id1 : dataFiles.keySet()) {
			this.indexMap.put(id1, i);
			this.labelMap.put(i, id1);
			String inputFile = dataFiles.get(id1);
			SequenceLabeller te = new SequenceLabeller();
			
			// set the special options to the TermExtractor
			te.setEvalPeriod(100);
			te.setNFolds(0); // no cross-validation
			te.setIterations(Integer.MAX_VALUE);
			te.setTrainingFraction(1.0);
			te.setRandomSeed(0);
			te.setGaussianVariance(10.0);
			//te.setDebugPipes(true);
			
			// train model
			try {
				System.out.println("[CLUSTERING] Training CRF on data file: " + inputFile + " (" + (i + 1) + " of " + this.dataFiles.keySet().size() + ")");
				te.trainModel(inputFile, numThreads);
				te.saveModel("trainedModels/" + id1 + ".model.gz");
			} catch (IOException e) {
				e.printStackTrace();
			} catch (Exception e) {
				e.printStackTrace();
			}
			
			// apply model to other data instances and determine log-likelihood
			int j = 0;
			for (String id2 : this.dataFiles.keySet()) {
				String instance = this.dataFiles.get(id2);
				System.out.println("[CLUSTERING] Applying CRF from " + id1 + " to " + id2 + " (" + (j + 1) + " of " + this.dataFiles.keySet().size() + ")");
				try {
					double[] labelProbabilities = te.labelProbabilities(instance);
					double logScore = labelProbabilities[0];
					double logZ = labelProbabilities[1];
					double prob = labelProbabilities[2];
					double loglikelihood = te.loglikelihood(instance);
					this.logScoreMatrix[i][j] = logScore;
					this.logZMatrix[i][j] = logZ;
					this.probMatrix[i][j] = prob;
					this.loglikelihoodMatrix[i][j] = loglikelihood;
					System.out.println("[CLUSTERING] 	logScore=" + logScore);
					System.out.println("[CLUSTERING] 	logZ=" + logZ);
					System.out.println("[CLUSTERING] 	prob=" + prob);
					System.out.println("[CLUSTERING] 	avg loglikelihood=" + loglikelihood);
				} catch (IOException e) {
					e.printStackTrace();
				}
				j++;
			}
			
			i++;
		}
		writeMatrix(logScoreMatrix, dataFiles.keySet(), "logScore_matrix.txt");
		writeMatrix(logZMatrix, dataFiles.keySet(), "logZ_matrix.txt");
		writeMatrix(probMatrix, dataFiles.keySet(), "prob_matrix.txt");
		writeMatrix(loglikelihoodMatrix, dataFiles.keySet(), "likelihood_matrix.txt");
		return loglikelihoodMatrix;
	}
	
	public HNode buildDendrogram(int linkage) {
		System.out.println("[CLUSTERING] Number of violations of the triangle inequity in the distance matrix: " + DistanceMetrics.checkTriangleInequity(this.loglikelihoodMatrix));
		System.out.println("[CLUSTERING] Building dendrogram...");
		System.out.println("[CLUSTERING] Num instances: " + this.numInstances);
		// Initialize unclustered instances
		ArrayList<HNode> hnodes = new ArrayList<HNode>();
		for (int i : this.labelMap.keySet()) {
			hnodes.add(new HNode(i, labelMap));
		}
		
		// Make log-likelihood matrix positive and symmetric, i.e. m[i][j] = -m[i][j]
		for (int i=0; i<this.numInstances; i++) {
			for (int j=i; j<this.numInstances; j++) {
				this.loglikelihoodMatrix[i][j] = -1.0 * ((this.loglikelihoodMatrix[i][j] + this.loglikelihoodMatrix[j][i]) / 2);
				this.loglikelihoodMatrix[j][i] = this.loglikelihoodMatrix[i][j];
				//System.out.print(this.loglikelihoodMatrix[i][j] + " ");
			}
			//System.out.println();
		}
		
		// Make tree
		while (hnodes.size() != 1) {
			// Find best pair of clusters to merge
			HNode bestMerge1 = null;
			HNode bestMerge2 = null;
			double bestMergeScore = Double.MAX_VALUE;
			for (HNode hnode1 : hnodes) {
				for (HNode hnode2 : hnodes) {
					double score = 0;
					switch (linkage) {
						case CRFHierarchicalClustering.LINKAGE_SINGLE: score = distanceSingle(hnode1, hnode2); break;
						case CRFHierarchicalClustering.LINKAGE_COMPLETE: score = distanceComplete(hnode1, hnode2); break;
						case CRFHierarchicalClustering.LINKAGE_UPGMA: score = distanceUPGMA(hnode1, hnode2); break;
						case CRFHierarchicalClustering.LINKAGE_WPGMA: score = distanceWPGMA(hnode1, hnode2); break;
						case CRFHierarchicalClustering.LINKAGE_WARD: score = distanceWard(hnode1, hnode2); break;
						default: score = distanceSingle(hnode1, hnode2); break;
					}
					if (hnode1.getNodeId() != hnode2.getNodeId() && score < bestMergeScore) {
						bestMergeScore = score;
						bestMerge1 = hnode1;
						bestMerge2 = hnode2;
					}
				}
			}
			
			// Merge clusters
			//System.out.println("[CLUSTERING] Merging clusters: " + bestMerge1 + " and " + bestMerge2 + " score=" + bestMergeScore);
			HNode newNode = new HNode(bestMerge1, bestMerge2, labelMap, bestMergeScore);
			hnodes.remove(bestMerge1);
			hnodes.remove(bestMerge2);
			hnodes.add(newNode);
		}
		
		this.root = hnodes.get(0);
		return root; // return root
	}
	
	/**
	 * Assigns dist(u,v) = MIN(dist(u[i],v[j]) for all points i in cluster u and j in cluster v.
	 * find single link distance aka minimum link, which is the closest distance between
     * any item in cluster1 and any item in cluster2
     * 
	 * @param hnode1
	 * @param hnode2
	 * @return
	 */
	private double distanceSingle(HNode hnode1, HNode hnode2) {
		double minLogLikelihood = Double.MAX_VALUE;
		for (int i1 : hnode1.getInstances()) {
			for (int i2 : hnode2.getInstances()) {
				double loglikelihood = this.loglikelihoodMatrix[i1][i2];
				//System.out.println("Debug distance: "+labelMap.get(i1)+"-"+labelMap.get(i2)+"="+loglikelihood);
				if (loglikelihood < minLogLikelihood) {
					minLogLikelihood = loglikelihood;
				}
			}
		}
		return minLogLikelihood;
	}
	
	/**
	 * Assigns dist(u,v) = MAX(dist(u[i],v[j]) for all points i in cluster u and j in cluster v.
	 * find complete link distance aka maximum link, which is the largest distance between
     * any item in cluster1 and any item in cluster2
     * 
	 * @param hnode1
	 * @param hnode2
	 * @return
	 */
	private double distanceComplete(HNode hnode1, HNode hnode2) {
		double maxLogLikelihood = Double.MIN_VALUE;
		for (int i1 : hnode1.getInstances()) {
			for (int i2 : hnode2.getInstances()) {
				double loglikelihood = this.loglikelihoodMatrix[i1][i2];
				if (loglikelihood > maxLogLikelihood) {
					maxLogLikelihood = loglikelihood;
				}
			}
		}
		return maxLogLikelihood;
	}
	
	/**
	 * UPGMA distance a.k.a. AVERAGE
	 * Assigns dist(u,v) = \sum_{ij} { dist(u[i], v[j]) } / (|u|*|v|) for all points i and j 
	 * where |u| and |v| are the cardinalities of clusters u and v, respectively.
	 * finds average distance between the elements of the two clusters
     *
	 * @param hnode1
	 * @param hnode2
	 * @return
	 */
	private double distanceUPGMA(HNode hnode1, HNode hnode2) {
		double sumLogLikelihood = 0;
		for (int i : hnode1.getInstances()) {
			for (int j : hnode2.getInstances()) {
				double loglikelihood = this.loglikelihoodMatrix[i][j];
				sumLogLikelihood += loglikelihood;
			}
		}
		return (sumLogLikelihood / (hnode1.getSize() * hnode2.getSize()));
	}
	
	/**
	 * WPGMA distance a.k.a. MEAN
	 * calculates the mean distance of a merged cluster (akak Group-average agglomerative clustering)
     *
	 * @param hnode1
	 * @param hnode2
	 * @return
	 */
	private double distanceWPGMA(HNode hnode1, HNode hnode2) {
		// calculates the mean distance of a merged cluster (akak Group-average
		// agglomerative clustering)
		HNode merged = new HNode(hnode1, hnode2, labelMap, 0);
		double fBestDist = 0;
		for (int i = 0; i < merged.getSize(); i++) {
			int i1 = merged.getInstances().get(i);
			for (int j = i + 1; j < merged.getSize(); j++) {
				int i2 = merged.getInstances().get(j);
				fBestDist += this.loglikelihoodMatrix[i1][i2];
			}
		}
		int n = merged.getSize();
		fBestDist /= (n * (n - 1.0) / 2.0);
		return fBestDist;
	}
	
	/**
	 * Uses the Ward variance minimization algorithm.
	 * finds the distance of the change in caused by merging the cluster.
     * The information of a cluster is calculated as the error sum of squares of the
     * centroids of the cluster and its members.
     *
	 * @param hnode1
	 * @param hnode2
	 * @return
	 */
	private double distanceWard(HNode hnode1, HNode hnode2) {
		// finds the distance of the change in caused by merging the cluster.
		// The information of a cluster is calculated as the error sum of
		// squares of the
		// centroids of the cluster and its members.
		double ESS1 = calcESS(hnode1);
		double ESS2 = calcESS(hnode2);
		HNode merged = new HNode(hnode1, hnode2, labelMap, 0);
		double ESS = calcESS(merged);
		double fBestDist = ESS * merged.getSize() - ESS1 * hnode1.getSize()
				- ESS2 * hnode2.getSize();
		return fBestDist;
	}
	
	/**
	 * Calculates error sum-of-squares for instances wrt centroid
	 * 
	 * @param cluster
	 * @return
	 */
	double calcESS(HNode hnode) {
		double[] centroid = new double[this.loglikelihoodMatrix.length];
		for (int instance : hnode.getInstances()) {
			for (int j = 0; j < this.loglikelihoodMatrix.length; j++) {
				centroid[j] += this.loglikelihoodMatrix[instance][j];
			}
		}
		for (int j = 0; j < this.loglikelihoodMatrix.length; j++) {
			centroid[j] /= hnode.getSize();
		}
		double fESS = 0;
		for (int instance : hnode.getInstances()) {
			fESS += euclideanDistance(centroid, this.loglikelihoodMatrix[instance]);
		}
		return fESS / hnode.getSize();
	} 
	
	public double euclideanDistance(double[] point1, double[] point2) { //debug method pls remove after use
		double distance = 0;
		for (int i = 0; i < point1.length; i++) {
			distance = distance + Math.pow((point1[i] - point2[i]), 2);
		}
		return Math.sqrt(distance);
	}
	
	public void printDendrogram() {
		new DendrogramPrinter(labelMap).print(root);
	}
	
	public ArrayList<HCluster> prune(int numClusters) {
		System.out.println("[CLUSTERING] Pruning tree: numClusters=" + numClusters);
		if (numClusters > this.numInstances) {
			throw new RuntimeException("More clusters than instances!");
		}
		ArrayList<HCluster> clusters = new ArrayList<HCluster>();
		ArrayList<HNode> unexpandedNodes = new ArrayList<HNode>(Arrays.asList(this.root));
		while (unexpandedNodes.size() != numClusters) {
			// Find highest node from the unexpanded nodes that can be expanded
			HNode highestNode = null;
			for (HNode hnode : unexpandedNodes) {
				if (highestNode == null || hnode.getHeight() > highestNode.getHeight()) {
					if (hnode.getInstances().size() > 1) { // if node can be expanded
						highestNode = hnode;
					}
				}
			}
			
			// Expand highest node
			unexpandedNodes.remove(highestNode);
//			System.out.println("Adding left node of: " + highestNode);
//			System.out.println("\twhich is: " + highestNode.getLeft());
//			System.out.println("Adding right node of: " + highestNode);
//			System.out.println("\twhich is: " + highestNode.getRight());
//			System.out.println();
			unexpandedNodes.add(highestNode.getLeft());
			unexpandedNodes.add(highestNode.getRight());
		}
		
		for (HNode hnode : unexpandedNodes) {
			ArrayList<String> instances = new ArrayList<String>();
			for (int instance : hnode.getInstances()) {
				instances.add(labelMap.get(instance));
			}
			HCluster cluster = new HCluster(instances);
			System.out.println("[CLUSTERING] 	" + cluster);
			clusters.add(cluster);
		}
		return clusters;
	}
	
	public ArrayList<HCluster> pruneUsingSilhouette() {
		System.out.println("[CLUSTERING] Prunning dendrogram using silhouette...");
		double[] silhouettes = new double[this.numInstances];
		double bestSilhouette = -1;
		ArrayList<HCluster> bestClusters = null;
		for (int k=1; k<this.numInstances + 1; k++) {
			System.out.println("[CLUSTERING] Trying k=" + k);
			ArrayList<HCluster> clusters = prune(k);
			double silhouette = silhouette(clusters);
			silhouettes[k-1] = silhouette;
			if (bestClusters == null || silhouette > bestSilhouette) {
				if (k != 1 && k != this.numInstances) {
					bestSilhouette = silhouette;
					bestClusters = clusters;
				}
			}
		}
		System.out.println("[CLUSTERING] Silhouettes (k, silhouette): ");
		for (int k=1; k<this.numInstances + 1; k++) {
			System.out.println("[CLUSTERING] 	" + k + ", " + silhouettes[k-1]);
		}
		System.out.println("[CLUSTERING] Best silhouette=" + bestSilhouette + " for k=" + bestClusters.size());
		return bestClusters;
	}
	
	public ArrayList<HCluster> pruneByJump() {
		System.out.println("[CLUSTERING] Prunning dendrogram using jumps in height...");
		ArrayList<HNode> unexpandedNodes = new ArrayList<HNode>(Arrays.asList(this.root));
		double[] heights = new double[this.numInstances+1];
		int k = 0;
		while (unexpandedNodes.size() != this.numInstances) { // until all node have been expanded
			// Find highest node from the unexpanded nodes
			HNode highestNode = null;
			for (HNode hnode : unexpandedNodes) {
				if (highestNode == null || hnode.getHeight() > highestNode.getHeight()) {
					highestNode = hnode;
				}
			}
			heights[k] = highestNode.getHeight();
			k++;
			
			// Expand highest node
			unexpandedNodes.remove(highestNode);
			unexpandedNodes.add(highestNode.getLeft());
			unexpandedNodes.add(highestNode.getRight());
		}

		System.out.println("[CLUSTERING] Heights (k, height):");
		for (k=0; k<heights.length; k++) {
			System.out.println("[CLUSTERING] 	" + (k+1) + ", " + heights[k]);
		}

		System.out.println("[CLUSTERING] Height differences (k, height difference):");
		double[] heightDifferences = new double[this.numInstances];
		for (k=0; k<heights.length-2; k++) {
			heightDifferences[k] = heights[k] - heights[k+1];
			System.out.println("[CLUSTERING] 	" + (k+2) + ", " + heightDifferences[k]);
		}
		
		int bestK = 0;
		double greatestJump = Double.MIN_VALUE;
		for (k=1; k<heightDifferences.length-1; k++) { // don't consider k=1 and k=numInstances
			if (heightDifferences[k] > greatestJump) {
				greatestJump = heightDifferences[k];
				bestK = k+1;
			}
		}
		System.out.println("[CLUSTERING] Best k=" + bestK + " with height_difference=" + greatestJump);
		return prune(bestK);
	}
	
	public ArrayList<HCluster> pruneUsingDunnIndex() {
		System.out.println("[CLUSTERING] Prunning dendrogram using dunn index...");
		double[] dunnIndexes = new double[this.numInstances];
		double bestDunnIndex = -1;
		ArrayList<HCluster> bestClusters = null;
		for (int k=1; k<this.numInstances + 1; k++) {
			System.out.println("[CLUSTERING] Trying k=" + k);
			ArrayList<HCluster> clusters = prune(k);
			double dunnIndex = dunnIndex(clusters);
			dunnIndexes[k-1] = dunnIndex;
			if (bestClusters == null || dunnIndex > bestDunnIndex) {
				if (k != 1 && k != this.numInstances) {
					bestDunnIndex = dunnIndex;
					bestClusters = clusters;
				}
			}
		}
		System.out.println("[CLUSTERING] Dunn Indexes (k, dunn index): ");
		for (int k=1; k<this.numInstances + 1; k++) {
			System.out.println("[CLUSTERING] 	" + k + ", " + dunnIndexes[k-1]);
		}
		System.out.println("[CLUSTERING] Best dunn index=" + bestDunnIndex + " for k=" + bestClusters.size());
		return bestClusters;
	}
	
	public ArrayList<HCluster> pruneUsingHubertGamma() {
		System.out.println("[CLUSTERING] Prunning dendrogram using Hubert gamma statistic...");
		double[] hubertGammas = new double[this.numInstances];
		double bestHubertGamma = -1;
		ArrayList<HCluster> bestClusters = null;
		for (int k=1; k<this.numInstances + 1; k++) {
			System.out.println("[CLUSTERING] Trying k=" + k);
			ArrayList<HCluster> clusters = prune(k);
			double hubertGamma = hubertGammaStatistic(clusters);
			hubertGammas[k-1] = hubertGamma;
			if (bestClusters == null || hubertGamma > bestHubertGamma) {
				if (k != 1 && k != this.numInstances) {
					bestHubertGamma = hubertGamma;
					bestClusters = clusters;
				}
			}
		}
		System.out.println("[CLUSTERING] Hubert gamma statistics (k, Hubert gamma): ");
		for (int k=1; k<this.numInstances + 1; k++) {
			System.out.println("[CLUSTERING] 	" + k + ", " + hubertGammas[k-1]);
		}
		System.out.println("[CLUSTERING] Best Hubert gamma=" + bestHubertGamma + " for k=" + bestClusters.size());
		return bestClusters;
	}
	
	public double silhouette(ArrayList<HCluster> clusters) {
		System.out.println("[CLUSTERING] Determining silhouette...");
		PrintWriter pw = null;
		try {
			pw = new PrintWriter("silhouettes.txt");
		} catch (FileNotFoundException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}
		double sumSiDataSet = 0;
		for (HCluster cluster : clusters) {
			double sumSiCluster = 0;
			for (String instance : cluster.getInstances()) {
				int instanceIndex = this.indexMap.get(instance);
				
				// Determine a(i) - let a(i) be the average dissimilarity of i with all other data within the same cluster.
				double ai = 0;
				if (cluster.getSize() == 1) {
					ai = 0;
				} else {
					for (String instance2 : cluster.getInstances()) {
						int instanceIndex2 = this.indexMap.get(instance2);
						if (instanceIndex != instanceIndex2) {
							ai += loglikelihoodMatrix[instanceIndex][instanceIndex2];
						}
					}
					ai = ai / (cluster.getSize() - 1);
				}
				
				// Determine b(i) - Then find the average dissimilarity of i with the data of another single cluster. 
				// Repeat this for every cluster of which i is not a member. 
				// Denote the lowest average dissimilarity to i of any such cluster by b(i).
				double lowestBi = Double.MAX_VALUE;
				for (HCluster cluster2 : clusters) {
					if (cluster2.getClusterId() != cluster.getClusterId()) {
						double bi = 0;
						for (String instance2 : cluster2.getInstances()) {
							int instanceIndex2 = this.indexMap.get(instance2);
							bi += loglikelihoodMatrix[instanceIndex][instanceIndex2];
						}
						bi = bi / cluster2.getSize();
						if (bi < lowestBi) {
							lowestBi = bi;
						}
					}
				}
				
				// Determining s(i)
				System.out.println("[CLUSTERING]       a("+instance+")=" + ai);
				System.out.println("[CLUSTERING]       b("+instance+")=" + lowestBi);
				double maxAB = Math.max(ai, lowestBi);
				double si = (lowestBi - ai) / Math.max(ai, lowestBi);
				if (maxAB <= 0) {
					si = 0;
				}
				pw.println(instance + "," + cluster.getClusterId() + "," + si);
				sumSiCluster += si;
				sumSiDataSet += si;
				System.out.println("[CLUSTERING]       S("+instance+")=" + si);
			}
			double avgSiCluster = sumSiCluster / cluster.getSize();
			System.out.println("[CLUSTERING]    Average S(i) for cluster " + cluster.getClusterId() + ": " + avgSiCluster);
		}
		double avgSiDataSet = sumSiDataSet / numInstances;
		System.out.println("[CLUSTERING] Average S(i) for dataset: " + avgSiDataSet);
		pw.close();
		return avgSiDataSet;
	}
	
	public double dunnIndex(ArrayList<HCluster> clusters) {
		// calculate the maximum diameter of a cluster 
		double maxDiam = Double.MIN_VALUE;
		for (HCluster cluster : clusters) {
			double diam = dunnDiam(cluster);
			if (diam > maxDiam) {
				maxDiam = diam;
			}
		}
		
		// calculate the minimum of the division: d(c1,cj)/maxDiam
		double minValue = Double.MAX_VALUE;
		for (int i=0; i<clusters.size(); i++) {
			HCluster cluster1 = clusters.get(i);
			for (int j=i+1; j<clusters.size(); j++) {
				HCluster cluster2 = clusters.get(j);
				double value = dunnDist(cluster1, cluster2) / maxDiam;
				if (value < minValue) {
					minValue = value;
				}
			}
		}
		System.out.println("[CLUSTERING] Dunn index: " + minValue);
		return minValue;
	}
	
	private double dunnDist(HCluster cluster1, HCluster cluster2) {
		double minLogLikelihood = Double.MAX_VALUE;
		for (String instance1 : cluster1.getInstances()) {
			int i1 = indexMap.get(instance1);
			for (String instance2 : cluster2.getInstances()) {
				int i2 = indexMap.get(instance2);
				double loglikelihood = this.loglikelihoodMatrix[i1][i2];
				//System.out.println("Debug distance: "+labelMap.get(i1)+"-"+labelMap.get(i2)+"="+loglikelihood);
				if (loglikelihood < minLogLikelihood) {
					minLogLikelihood = loglikelihood;
				}
			}
		}
		return minLogLikelihood;
	}
	
	private double dunnDiam(HCluster cluster) {
		double maxLogLikelihood = Double.MIN_VALUE;
		for (String instance1 : cluster.getInstances()) {
			int i1 = indexMap.get(instance1);
			for (String instance2 : cluster.getInstances()) {
				int i2 = indexMap.get(instance2);
				double loglikelihood = this.loglikelihoodMatrix[i1][i2];
				if (loglikelihood > maxLogLikelihood) {
					maxLogLikelihood = loglikelihood;
				}
			}
		}
		return maxLogLikelihood;
	}
	
	public double hubertGammaStatistic(ArrayList<HCluster> clusters) {
		if (clusters.size() == 1 || clusters.size() == this.numInstances) {
			return Double.NaN; // Hubert gamma statistic is undefined for nc=1 and nc=N
		}
		
		// determine the Q matrix
		double[][] qMatrix = new double[this.numInstances][this.numInstances];
		for (int i=0; i<clusters.size(); i++) {
			HCluster cluster1 = clusters.get(i);
			for (int j=i; j<clusters.size(); j++) { 
				HCluster cluster2 = clusters.get(j);
				
				// determine q for all the instances in these two clusters
				double q = 0;
				if (i != j) {
					double avgLogLikelihood = 0;
					for (String instance1 : cluster1.getInstances()) {
						int i1 = indexMap.get(instance1); 
						for (String instance2 : cluster2.getInstances()) {
							int i2 = indexMap.get(instance2);
							double loglikelihood = this.loglikelihoodMatrix[i1][i2];
							avgLogLikelihood += loglikelihood;
						}
					}
					avgLogLikelihood = avgLogLikelihood / (cluster1.getSize() * cluster2.getSize());
					q = avgLogLikelihood;
				}
				
				// set the q values in the Q matrix 
				for (String instance1 : cluster1.getInstances()) {
					int i1 = indexMap.get(instance1); 
					for (String instance2 : cluster2.getInstances()) {
						int i2 = indexMap.get(instance2);
						qMatrix[i1][i2] = q;
						qMatrix[i2][i1] = q;
					}
				}
			}
		}
		
		double sum = 0;
		for (int i=0; i<this.numInstances-1; i++) {
			for (int j=i+1; j<this.numInstances; j++) {
				sum += this.loglikelihoodMatrix[i][j] * qMatrix[i][j];
			}
		}
		double m = (this.numInstances * (this.numInstances - 1) ) / 2;
		double hubertGammaStatistic = sum / m;
		System.out.println("[CLUSTERING] Hubert gamma statistic: " + hubertGammaStatistic);
		return hubertGammaStatistic;
	}
	
	public static void printClusterAssignments(ArrayList<HCluster> clusters) {
		System.out.println("[CLUSTERING] Cluster assignments:");
		for (HCluster cluster : clusters) {
			for (String id : cluster.getInstances()) {
				System.out.println("assignments['" + id + "'] = " + cluster.getClusterId());
			}
		}
	}
	
	private static void writeMatrix(double[][] matrix, Set<String> idsSet, String outputFile) {
		try {
			PrintWriter pw = new PrintWriter(outputFile);
			pw.print("\t");
			String[] ids = new String[idsSet.size()];
			int i = 0;
			for (String id : idsSet) {
				ids[i] = id;
				i++;
				pw.print(id + "\t");
			}
			pw.println();
			for (i=0; i<matrix.length; i++) {
				pw.print(ids[i] + "\t");
				for (int j=0; j<matrix[0].length; j++) {
					pw.print(matrix[i][j] + "\t");
				}
				pw.println();
			}
			pw.flush();
			pw.close();
		} catch (FileNotFoundException e) {
			e.printStackTrace();
		}
	}
	
	public static void saveModel(CRFHierarchicalClustering clustering, String outputFile) {
		System.out.println("[CLUSTERING] Saving model to file: " + outputFile);
		FileOutputStream fOut;
		try {
			fOut = new FileOutputStream(outputFile);
			ObjectOutputStream objOut = new ObjectOutputStream (fOut);
			objOut.writeObject(clustering);
			fOut.close();
		} catch (Exception e) {
			e.printStackTrace();
		}
	}
	
	public static CRFHierarchicalClustering loadModel(String inputFile) {
		System.out.println("[CLUSTERING] Loading model from file: " + inputFile);
		FileInputStream fIn;
		try {
			fIn = new FileInputStream(inputFile);
			ObjectInputStream objIn = new ObjectInputStream(fIn);
			Object obj = objIn.readObject();

			if (obj instanceof CRFHierarchicalClustering) {
				CRFHierarchicalClustering clustering = (CRFHierarchicalClustering) obj;
				return clustering;
			} else {
				throw new RuntimeException("Input file is not a valid CRFHierarchicalClustering object: " + inputFile);
			}
		} catch (Exception e) {
			e.printStackTrace();
		}
		return null;
	}
	
	public static ArrayList<HCluster> loadClustersFromFile(String inputFile) {
		ArrayList<HCluster> clusters = new ArrayList<HCluster>();
		try {
			BufferedReader br = new BufferedReader(new FileReader(inputFile));
			String line = null;
			while((line = br.readLine()) != null) {
				ArrayList<String> instances = new ArrayList<String>();
				line = line.replaceAll("Cluster [0-9]+: ", "");
				String[] splitedLine = line.split(",");
				for (String instance : splitedLine) {
					instance = instance.trim();
					if (instance.length() > 0) {
						instances.add(instance);
					}
				}
				clusters.add(new HCluster(instances));
			}
		} catch (IOException e) {
			e.printStackTrace();
		}
		return clusters;
	}
	
	public static void main(String[] args) { 
//		int numThreads = 4;
//		//CRFHierarchicalClustering clustering = new CRFHierarchicalClustering("data/synthetic_data_v2/stddev0.5", numThreads);
//		CRFHierarchicalClustering clustering = new CRFHierarchicalClustering("data/mturk43", numThreads);
//		clustering.calcLogLikelihoods();
//		//CRFHierarchicalClustering.saveModel(clustering, "hcluster_synthetic_stddev0.5.model");
//		CRFHierarchicalClustering.saveModel(clustering, "hcluster_mturk.model");

//		CRFHierarchicalClustering clustering = CRFHierarchicalClustering.loadModel("clustering_results/" +
//				"synthetic_data_v2/stddev0.2/hcluster_synthetic_stddev0.2.model");
		CRFHierarchicalClustering clustering = CRFHierarchicalClustering.loadModel("clustering_results/" +
			"mturk43/hcluster_mturk.model");
		clustering.buildDendrogram(CRFHierarchicalClustering.LINKAGE_UPGMA);
		clustering.printDendrogram();
		ArrayList<HCluster> clusters = clustering.prune(3);
		clustering.silhouette(clusters);
		clustering.dunnIndex(clusters);
		clustering.hubertGammaStatistic(clusters);
//		ArrayList<HCluster> clusters = clustering.pruneUsingSilhouette();
		//ArrayList<HCluster> clusters = clustering.pruneUsingDunnIndex();
		//ArrayList<HCluster> clusters = clustering.pruneUsingHubertGamma();
		//ArrayList<HCluster> clusters = clustering.pruneByJump();
		CRFHierarchicalClustering.printClusterAssignments(clusters);

//		ArrayList<HCluster> clusters2 = CRFHierarchicalClustering.loadClustersFromFile("clustering_results/kclusters.txt");
//		for (HCluster c : clusters2) {
//			System.out.println("CLUSTER: " + c);
//		}
//		clustering.silhouette(clusters2);
	}
}

