package pt.uc.dei.crf_ma.pipe;

import java.io.IOException;
import java.io.InputStream;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.io.Serializable;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.regex.Pattern;

import opennlp.tools.sentdetect.SentenceDetectorME;
import opennlp.tools.sentdetect.SentenceModel;
import opennlp.tools.tokenize.TokenizerME;
import opennlp.tools.tokenize.TokenizerModel;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import pt.uc.dei.crf_ma.SequenceLabeller;
import cc.mallet.pipe.Pipe;
import cc.mallet.types.Instance;
import cc.mallet.types.LabelAlphabet;
import cc.mallet.types.LabelSequence;
import cc.mallet.types.Token;
import cc.mallet.types.TokenSequence;

public class AnnotatedText2TokenSequence extends Pipe implements Serializable {

	// Serialization garbage
	private static final long serialVersionUID = 1;
	private Logger logger = LoggerFactory.getLogger(AnnotatedText2TokenSequence.class);
	
	public static String START_SENTENCE_FEATURE = "START_SENTENCE";
	public static String START_PARAGRAPH_FEATURE = "START_PARAGRAPH"; 
	public static String SENTENCE_NUMBER_FEATURE = "SENTENCE_NUMBER=";
	public static String PARAGRAPH_NUMBER_FEATURE = "PARAGRAPH_NUMBER=";
	public static String DEFAULT_TOKENIZER_MODEL_FILE = "/models/opennlp/en-token.bin";
	public static String DEFAULT_SENTENCE_DETECTOR_MODEL_FILE = "/models/opennlp/en-sent.bin"; 
	private boolean setTokensAsFeatures;
	private String startTagPattern = "\\[";
	private String endTagPattern = "\\]";
	// OpneNLP Models
	private TokenizerME tokenizer;
	private SentenceDetectorME sentenceDetector;

	/**
	 * Creates a new <code>AnnotatedText2TokenSequence</code> instance. By
	 * default we include tokens as features.
	 */
	public AnnotatedText2TokenSequence() {
		super(null, new LabelAlphabet());
		this.setTokensAsFeatures = true;
		loadOpenNLPModels(DEFAULT_TOKENIZER_MODEL_FILE, DEFAULT_SENTENCE_DETECTOR_MODEL_FILE);
	}

	/**
	 * creates a new <code>AnnotatedText2TokenSequence</code> instance which
	 * includes tokens as features iff the supplied argument is true.
	 */
	public AnnotatedText2TokenSequence(boolean inc) {
		super(null, new LabelAlphabet());
		this.setTokensAsFeatures = inc;
		loadOpenNLPModels(DEFAULT_TOKENIZER_MODEL_FILE, DEFAULT_SENTENCE_DETECTOR_MODEL_FILE);
	}

	/**
	 * creates a new <code>AnnotatedText2TokenSequence</code> instance which
	 * includes tokens as features iff the supplied argument is true. Also
	 * receives the tag delimiters as parameters.
	 */
	public AnnotatedText2TokenSequence(boolean inc, String startTagPattern, String endTagPattern) {
		super(null, new LabelAlphabet());
		this.setTokensAsFeatures = inc;
		this.startTagPattern = startTagPattern;
		this.endTagPattern = endTagPattern;
		loadOpenNLPModels(DEFAULT_TOKENIZER_MODEL_FILE, DEFAULT_SENTENCE_DETECTOR_MODEL_FILE);
	}

	public Instance pipe(Instance carrier) {
		Object inputData = carrier.getData();
		if (!(inputData instanceof String)) {
			throw new IllegalArgumentException("Not a String; got " + inputData);
		}

		/*
		 * inputList format: each entry of the external list represents a token; 
		 * each entry of the internal list follows the following format:
		 * label, token, feature1, feature2, ..., featureN 
		 */
		ArrayList<ArrayList<String>> inputList = new ArrayList<ArrayList<String>>();
		String[] paragraphs = ((String) inputData).split("\n+");
		for (int p=0; p<paragraphs.length; p++) {
			boolean newParagraph = true;
			
			// use OpenNLP to split the paragraph in sentences
			String sentences[] = sentenceDetector.sentDetect(paragraphs[p]);
			
			for (int s=0; s<sentences.length; ) {
				boolean newSentence = true;
				
				
				int openTagsInSentence = 0;
				String sentence = "";
				String tokens[] = null;
				do {
					// keep merging sentences until open/close tags are balanced
					if (openTagsInSentence!=0 && s == sentences.length) {
						throw new RuntimeException("Unbalanced square brackets in the annotated sentence: " + sentence);
					}
					
					sentence += sentences[s++];
					
					// line preprocessing. these are the ones we want to separate
					// from adjacent words for sure.
					sentence = sentence.replaceAll("\\[", " [ ");
					sentence = sentence.replaceAll("\\]", " ] ");
					sentence = sentence.replaceAll("\\(", " ( ");
					sentence = sentence.replaceAll("\\)", " ) ");
					sentence = sentence.replaceAll(",", " , ");
					sentence = sentence.replaceAll("\\s+", " ");

					// apply OpneNLP tokenizer
					tokens = tokenizer.tokenize(sentence);
				} while ((openTagsInSentence = checkTagSyntax(tokens)) != 0);

				int openTags = 0;
				boolean first = true;
				for (String token : tokens) {
					if (Pattern.matches(startTagPattern, token)) {
						openTags++;
						first = true;
					} else if (Pattern.matches(endTagPattern, token)) {
						openTags--;
					} else {
						ArrayList<String> inputLine = null;
						if (openTags == 0) {
							inputLine = new ArrayList<String>(Arrays.asList(SequenceLabeller.DEFAULT_LABEL, token));
						} else if (openTags > 0) {
							if (first) {
								inputLine = new ArrayList<String>(Arrays.asList(SequenceLabeller.START_TAGS[0], token));
								first = false;
							} else {
								inputLine = new ArrayList<String>(Arrays.asList(SequenceLabeller.CONTINUE_TAGS[0], token));
							}
						} else {
							// this should never happen.. but if it happens drop the line and log it as SEVERE.
							logger.error("Unbalanced square brackets in the annotated sentence: {}", sentence);
							break;
							//throw new RuntimeException("Unbalanced square brackets in the annotated sentence: " + sentence);
						}
						inputList.add(inputLine);
						
						// add extra features to the token
						if(newParagraph) {
							inputLine.add(AnnotatedText2TokenSequence.START_PARAGRAPH_FEATURE);
							newParagraph = false;
						}
						if(newSentence) {
							inputLine.add(AnnotatedText2TokenSequence.START_SENTENCE_FEATURE);
							newSentence = false;
						}
						inputLine.add(AnnotatedText2TokenSequence.PARAGRAPH_NUMBER_FEATURE + (p+1));
						inputLine.add(AnnotatedText2TokenSequence.SENTENCE_NUMBER_FEATURE + (s+1));
					}
				}
			} 
		}

		LabelAlphabet labels;
		LabelSequence target = null;
		TokenSequence ts = new TokenSequence();

		labels = (LabelAlphabet) getTargetAlphabet();
		target = new LabelSequence(labels, inputList.size());
			
		for (int l = 0; l < inputList.size(); l++) {
			// inputLine format is: label, token, feature1, feature2, ..., featureN 
			ArrayList<String> inputLine = inputList.get(l);
			if (isTargetProcessing()) {
				if (inputLine.size() < 1) {
					throw new IllegalStateException("Missing label at line "
							+ l + " instance " + carrier.getName());
				}
				target.add(inputLine.get(0));
			}
			Token tok = new Token(inputLine.get(1));
			if (setTokensAsFeatures) {
				tok.setFeatureValue(inputLine.get(1), 1.0);
			}
			for(int fnum = 2; fnum < inputLine.size(); fnum++) {
				tok.setFeatureValue(inputLine.get(fnum), 1.0);
			}
			ts.add(tok);
		}
		
		carrier.setData(ts);
		carrier.setTarget(target);
		
		return carrier;
	}
	
	private int checkTagSyntax(String[] tokens) {
		int openTags = 0;
		for (String token : tokens) {
			if (Pattern.matches(startTagPattern, token)) {
				openTags++;
			} else if (Pattern.matches(endTagPattern, token)) {
				openTags--;
			} 
		}
		return openTags;
	}

	private void loadOpenNLPModels(String tokenizerModelFile, String sentenceDetectorModelFile) {
		InputStream modelIn = null;
		try {
			modelIn = getClass().getResourceAsStream(tokenizerModelFile);
			TokenizerModel model = new TokenizerModel(modelIn);
			this.tokenizer = new TokenizerME(model);
		} catch (IOException e) {
			e.printStackTrace();
		} finally {
			if (modelIn != null) {
				try {
					modelIn.close();
				} catch (IOException e) {
				}
			}
		}
		try {
			modelIn = getClass().getResourceAsStream(sentenceDetectorModelFile);
			SentenceModel model = new SentenceModel(modelIn);
			this.sentenceDetector = new SentenceDetectorME(model);
		} catch (IOException e) {
			e.printStackTrace();
		} finally {
			if (modelIn != null) {
				try {
					modelIn.close();
				} catch (IOException e) {
				}
			}
		}
	}


	private void writeObject(ObjectOutputStream out) throws IOException {
		out.writeLong(this.serialVersionUID);
		out.writeBoolean(setTokensAsFeatures);
		out.writeObject(startTagPattern);
		out.writeObject(endTagPattern);
	}

	private void readObject(ObjectInputStream in) throws IOException, ClassNotFoundException {
		long version = in.readLong();
		setTokensAsFeatures = in.readBoolean();
		startTagPattern = (String) in.readObject();
		endTagPattern = (String) in.readObject();
		
		//reload logger
		logger = LoggerFactory.getLogger(AnnotatedText2TokenSequence.class);
		loadOpenNLPModels(DEFAULT_TOKENIZER_MODEL_FILE, DEFAULT_SENTENCE_DETECTOR_MODEL_FILE);
	}
}
