package pt.uc.dei.crf_ma.pipe;

import java.io.BufferedReader;
import java.io.File;
import java.io.FileNotFoundException;
import java.io.FileReader;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.LineNumberReader;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.io.Reader;
import java.io.Serializable;
import java.util.HashMap;
import java.util.HashSet;
import java.util.zip.GZIPInputStream;

import cc.mallet.pipe.Pipe;
import cc.mallet.types.Instance;
import cc.mallet.types.Token;
import cc.mallet.types.TokenSequence;

public class DbpediaLexiconMembership extends Pipe implements Serializable
{
	public static int MAX_SEARCH_LENGTH = 5;
	private String name;
	private String lexiconPath;
	private HashMap<String, Double> lexicon;
	boolean ignoreCase;
	
	public DbpediaLexiconMembership(String name, String lexiconPath, boolean ignoreCase) {
		this.name = name;
		this.lexiconPath = lexiconPath;
		this.ignoreCase = ignoreCase;
		this.lexicon = loadLexicon(lexiconPath);
	}
	
	private HashMap<String, Double> loadLexicon(String lexiconPath) {
		BufferedReader reader = null;
		if (lexiconPath.endsWith(".gz")) {
			try {
				InputStream modelFile = new GZIPInputStream(getClass().getResourceAsStream(lexiconPath));
				reader = new BufferedReader(new InputStreamReader(modelFile));
			} catch (IOException e) {
				e.printStackTrace();
			}
		} else if (lexiconPath.endsWith(".txt")) {
			InputStream modelFile = getClass().getResourceAsStream(lexiconPath);
			reader = new BufferedReader(new InputStreamReader(modelFile));
		}
		
		if(reader == null) {
			throw new RuntimeException("Unable to load lexicon from file: " + lexiconPath);
		}

		HashMap<String, Double> lexicon = new HashMap<String, Double>();
		String line = null;
		int count = 0;
		try {
			while ((line = reader.readLine()) != null) {
				if ((++count) % 100000 == 0) {
					System.out.println("loaded: " + count);
				}
				String[] splitedLine = line.split("\t");
				String name = splitedLine[1].trim();
				if (name.length() < 3) 
					continue; // ignore very small names
				if (ignoreCase)
					name = name.toLowerCase();
				Double pagerank = Double.parseDouble(splitedLine[2].trim());
				/*
				 * min:-17.8945
				 * max:-5.9503407
				 * count:10476611
				 * avg:-17.3181690272
				 */
				double min = -17.8945;
				double max = -5.9503407;
				pagerank = (pagerank - min) / (max - min);
				if (!lexicon.containsKey(name)) 
					lexicon.put(name, pagerank); 
			}
		} catch (IOException e) {
			e.printStackTrace();
		}
		
		if (lexicon.size() == 0) {
			throw new RuntimeException("Empty lexicon " + lexiconPath);
		}
		
		return lexicon;
	}
	
	public Instance pipe (Instance carrier)
	{
		TokenSequence ts = (TokenSequence) carrier.getData();
		for (int i = 0; i < ts.size(); i++) {
			for (int length = MAX_SEARCH_LENGTH; length > 0; length--) {
				String s = "";
				for (int j = 0; j < length && (i + j) < ts.size(); j++) {
					s += ts.get(i + j).getText() + " ";
				}
				if (ignoreCase) {
					s = s.toLowerCase();
				}
				s = s.trim();
				
				if (lexicon.containsKey(s)) {
					double pagerank = lexicon.get(s); 
					for (int j = 0; j < length && (i + j) < ts.size(); j++) {
						if (length == 1 || i == ts.size() - 1) {
							ts.get(i + j).setFeatureValue(name, pagerank);
						} else {
							ts.get(i + j).setFeatureValue("MULTI" + name, pagerank);
						}
					}
					break;
				}
			}
		}
		return carrier;
	}
	
	// Serialization 
	private static final long serialVersionUID = 1;
	private static final int CURRENT_SERIAL_VERSION = 0;
	
	private void writeObject (ObjectOutputStream out) throws IOException {
		out.writeInt(CURRENT_SERIAL_VERSION);
		out.writeObject(name);
		out.writeObject(lexiconPath);
		out.writeBoolean(ignoreCase);
	}
	
	private void readObject (ObjectInputStream in) throws IOException, ClassNotFoundException {
		int version = in.readInt();
		this.name = (String) in.readObject();
		this.lexiconPath = (String) in.readObject();
		this.ignoreCase = in.readBoolean();
		this.lexicon = loadLexicon(lexiconPath);
	}

	public String getName() {
		return name;
	}

	public void setName(String name) {
		this.name = name;
	}
}