package pt.uc.dei.crf_ma.pipe;

import java.io.BufferedReader;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.FileReader;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.HashMap;
import java.util.Iterator;
import java.util.regex.Pattern;
import java.util.zip.GZIPInputStream;

import opennlp.tools.sentdetect.SentenceDetectorME;
import opennlp.tools.sentdetect.SentenceModel;
import opennlp.tools.tokenize.TokenizerME;
import opennlp.tools.tokenize.TokenizerModel;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import pt.uc.dei.crf_ma.SequenceLabeller;
import pt.uc.dei.crf_ma.types.TermExtractorInstance;

import cc.mallet.types.Instance;

public class SentencesAsInstancesFileIterator implements Iterator<Instance> {
	
	// Serialization garbage
	private static final long serialVersionUID = 1;
	private Logger logger = LoggerFactory.getLogger(AnnotatedText2TokenSequence.class);
	
	// feature names
	//public static String START_SENTENCE_FEATURE = "START_SENTENCE";
	public static final String START_PARAGRAPH_FEATURE = "START_PARAGRAPH"; 
	public static final String RELATIVE_SENTENCE_NUMBER_FEATURE = "RELATIVE_SENTENCE_NUMBER";
	public static final String GLOBAL_SENTENCE_NUMBER_FEATURE = "GLOBAL_SENTENCE_NUMBER";
	public static final String PARAGRAPH_NUMBER_FEATURE = "PARAGRAPH_NUMBER";
	public static final String TFIDF_FEATURE = "TFIDF";
	public static final String TF_FEATURE = "TF";
	public static final String IDF_FEATURE = "IDF";
	
	// input files paths
	public static final String DEFAULT_TOKENIZER_MODEL_FILE = "/models/opennlp/en-token.bin";
	public static final String DEFAULT_SENTENCE_DETECTOR_MODEL_FILE = "/models/opennlp/en-sent.bin"; 
	public static final String IDFPATH = "/pipes/idf.gz";
	
	// OpneNLP Models
	private TokenizerME tokenizer;
	private SentenceDetectorME sentenceDetector;

	private String path;
	private String currentFile;
	private String validFilesRegex; // regex to identify what filename are valid
	private ArrayList<File> fileArray; // array of all files
	private Iterator<File> fileIterator; // iterator over the array of all files
	private ArrayList<Sentence> currentFileSentences; // array with all the sentences in a file
	private Iterator<Sentence> currentFileSentencesIterator; // iterator over the array with all sentences of a file
	private int currentSentenceParagraph; // the number of the current paragraph
	private int currentSentenceRelativeNumber; // sentence number in the paragraph
	private int currentSentenceGlobalNumber; // sentence number in the whole file
	private String instanceId;

	private String startTagPattern;
	private String endTagPattern;
	
	// tfidf related variables
	private boolean addTFIDFFeatures; //should we add a feature with the TFIDF to the instance?
	private HashMap<String, Double> inverseDocumentFrequencies;
	private HashMap<String, Integer> termFrequencies; // the numerators of the tf formula
	private int numTermsInDocument; // this is the denominator of the tf formula
	
	public SentencesAsInstancesFileIterator(String path, String validFilesRegex, String startTagPattern, String endTagPattern, boolean addTFIDFFeatures) {
		this.path = path;
		this.validFilesRegex = validFilesRegex;
		this.startTagPattern = startTagPattern;
		this.endTagPattern = endTagPattern;
		this.addTFIDFFeatures = addTFIDFFeatures;
		this.fileArray = new ArrayList<File>();
		
		File source = new File(path);
		if(!source.exists()) {
			throw new IllegalArgumentException("Invalid input file/directory: " + path);
		}
		if(source.isDirectory()) {
			File[] files = source.listFiles();
			if (files == null) {
			    throw new IllegalArgumentException("Directory does not exist or is empty: " + path);
			} else {
			    for (File file : files) {
			    	if (Pattern.matches(validFilesRegex, file.getName())) {
			    		fileArray.add(file);
			    	}
			    }
				if (fileArray.isEmpty()) {
					throw new IllegalArgumentException("Directory is empty: " + path);
				}
			}
		} else {
			fileArray.add(source);
		}
		this.fileIterator = fileArray.iterator();
		
		loadOpenNLPModels(DEFAULT_TOKENIZER_MODEL_FILE, DEFAULT_SENTENCE_DETECTOR_MODEL_FILE);
		if (addTFIDFFeatures) {
			try {
				inverseDocumentFrequencies = loadIDFMapping(IDFPATH);
			} catch (IOException e) {
				e.printStackTrace();
			}
		}
	}
	
	public boolean hasNext() {
		return fileIterator.hasNext() || currentFileSentencesIterator.hasNext();
	}

	public Instance next() {
		if (currentFileSentencesIterator == null || !currentFileSentencesIterator.hasNext()) {
			getNextFile(); // get a new file from the directory
		}

		// grab a new sentence from the sentence iterator; this will be the new instance to return
		Sentence sentence = currentFileSentencesIterator.next();
		boolean newParagraph = currentSentenceParagraph != sentence.getParagraphNumber();
		currentSentenceParagraph = sentence.getParagraphNumber();
		if (newParagraph) 
			currentSentenceRelativeNumber = 0;
		currentSentenceRelativeNumber++;
		currentSentenceGlobalNumber++;
		
		return buildInstance(sentence, newParagraph);
	}
	
	private void getNextFile() {
		File file = fileIterator.next();
		this.currentFile = file.getPath();
		//logger.info("Loading instances from file: " + file.getName());
		System.out.println("Loading instances from file: " + file.getName());
		BufferedReader reader = null;
		if (file.getName().endsWith(".gz")) {
			try {
				reader = new BufferedReader(new InputStreamReader(new GZIPInputStream(new FileInputStream(file))));
			} catch (FileNotFoundException e) {
				e.printStackTrace();
			} catch (IOException e) {
				e.printStackTrace();
			}
		} else {
			try {
				reader = new BufferedReader(new FileReader(file));
			} catch (FileNotFoundException e) {
				e.printStackTrace();
			}
		}
		
		// process new file
		int paragraphNumber = 0;
		int sentenceNumber = 0;
		this.currentFileSentences = new ArrayList<Sentence>();
		String line = null;
		try {
			String sentenceText = "";
			ArrayList<String> tokens = new ArrayList<String>();
			ArrayList<String> labels = new ArrayList<String>();
			while ((line = reader.readLine()) != null) {
				if (line.length() == 0) {
					if (tokens.size() > 0) {
						currentFileSentences.add(new Sentence(sentenceText, tokens, labels, paragraphNumber, sentenceNumber));
						sentenceNumber++;
					}
					sentenceText = "";
					tokens = new ArrayList<String>();
					labels = new ArrayList<String>();
					continue;
				}
				String[] split = line.split(" ");
				String token = split[0];
				String label = split[1];
				sentenceText += token + " ";
				tokens.add(token);
				labels.add(label);
			}
			if (tokens.size() > 0) {
				currentFileSentences.add(new Sentence(sentenceText, tokens, labels, paragraphNumber, sentenceNumber));
				sentenceNumber++;
			}
		} catch (IOException e) {
			e.printStackTrace();
		}

		try {
			reader.close();
		} catch (IOException e1) {
			e1.printStackTrace();
		}
		
		// it's a new file, so reset file related variable
		this.currentSentenceParagraph = 0;
		this.currentSentenceGlobalNumber = 0;
		this.instanceId = file.getName();
		this.termFrequencies = null;
		this.currentFileSentencesIterator = currentFileSentences.iterator();
	}
	
	private Instance buildInstance(Sentence sentence, boolean newParagraph) {
		if (addTFIDFFeatures && termFrequencies == null) {
			// calculate term frequencies
			// remember: tf is calculated over all sentences in the file
			termFrequencies = new HashMap<String, Integer>();
			numTermsInDocument = 0;
			for (Sentence sent : currentFileSentences) {
				for (int i=0; i<sent.getTokens().size(); i++) {
		        	String token = sent.getTokens().get(i).toLowerCase();
		        	numTermsInDocument++;
		        	if (termFrequencies.containsKey(token)) {
		        		Integer tf = termFrequencies.get(token);
		        		termFrequencies.put(token, new Integer(tf+1));
		        	} else {
		        		termFrequencies.put(token, new Integer(1));
		        	}
		        }
			}
		}
		
		StringBuffer sb = new StringBuffer();
		ArrayList<String> inputTokens = new ArrayList<String>();
		ArrayList<String> inputLabels = new ArrayList<String>();
		for (int i=0; i<sentence.getTokens().size(); i++) {
        	String token = sentence.getTokens().get(i);
        	String label = sentence.getLabels().get(i);
			inputTokens.add(token);
			sb.append(token).append(" ");
			// add extra features to the token
			if(newParagraph) {
				sb.append(SentencesAsInstancesFileIterator.START_PARAGRAPH_FEATURE).append(" ");
				newParagraph = false;
			}
			sb.append(SentencesAsInstancesFileIterator.PARAGRAPH_NUMBER_FEATURE).append("=").append(currentSentenceParagraph).append(" ");
			//sb.append(ArticleIterator.RELATIVE_SENTENCE_NUMBER_FEATURE).append("=").append(currentSentenceRelativeNumber).append(" ");
			sb.append(SentencesAsInstancesFileIterator.GLOBAL_SENTENCE_NUMBER_FEATURE).append("=").append(currentSentenceGlobalNumber).append(" ");
			if (addTFIDFFeatures) {
				Double idf = inverseDocumentFrequencies.get(token.toLowerCase());
//					if (idf == null) {
//						idf = 1.0;
//					}
				if (idf != null) {
					sb.append(SentencesAsInstancesFileIterator.IDF_FEATURE).append("=").append(idf).append(" ");
				}
				Integer occurencies = termFrequencies.get(token.toLowerCase());
				Double tf = null;
				if( occurencies != null) {
					tf = (double) termFrequencies.get(token.toLowerCase()).intValue() / (double) numTermsInDocument;
					sb.append(SentencesAsInstancesFileIterator.TF_FEATURE).append("=").append(tf).append(" ");
				}
				if (idf != null && occurencies != null) {
	        		double tfidf =  tf * idf.doubleValue();
	        		sb.append(SentencesAsInstancesFileIterator.TFIDF_FEATURE).append("=").append(tfidf).append(" ");
	        	} 
			}

			inputLabels.add(label);
			sb.append(label).append("\n");
		}
		
		//System.out.println("[MultipleTaggerDEBUG]" + sb.toString());
		
		TermExtractorInstance instance = new TermExtractorInstance(sb.toString(), null, "linegroup"+currentSentenceGlobalNumber, null);
		//logger.info("Article id: " + articleId);
		instance.setInstanceId(this.instanceId);
		instance.setTokens(inputTokens);
		instance.setLabels(inputLabels);
		
		//Instance result = new Instance(sb.toString(), null, "linegroup"+currentSentenceGlobalNumber, null);
		return (Instance) instance;
	}

	public void remove() {
		throw new IllegalStateException("This Iterator<Instance> does not support remove().");
	}
	
	private void loadOpenNLPModels(String tokenizerModelFile, String sentenceDetectorModelFile) {
		InputStream modelIn = null;
		try {
			modelIn = getClass().getResourceAsStream(tokenizerModelFile);
			TokenizerModel model = new TokenizerModel(modelIn);
			this.tokenizer = new TokenizerME(model);
		} catch (IOException e) {
			e.printStackTrace();
		} finally {
			if (modelIn != null) {
				try {
					modelIn.close();
				} catch (IOException e) {
				}
			}
		}
		try {
			modelIn = getClass().getResourceAsStream(sentenceDetectorModelFile);
			SentenceModel model = new SentenceModel(modelIn);
			this.sentenceDetector = new SentenceDetectorME(model);
		} catch (IOException e) {
			e.printStackTrace();
		} finally {
			if (modelIn != null) {
				try {
					modelIn.close();
				} catch (IOException e) {
				}
			}
		}
	}
	
	private HashMap<String, Double> loadIDFMapping(String IDFPATH) throws IOException {
		HashMap<String, Double> result = new HashMap<String, Double>();
		
		InputStream modelFile = new GZIPInputStream(getClass().getResourceAsStream(IDFPATH));
		BufferedReader reader = new BufferedReader(new InputStreamReader(modelFile));
		
		String line = null;
		while ((line = reader.readLine())!=null) {
			String [] fields = line.split("\t");
			String word = fields[0];
			double idf = Double.parseDouble(fields[1]);
			
			result.put(word.toLowerCase(), idf);
		}
		
		return result;
 	}
}

class Sentence {
	private String sentenceText;
	private ArrayList<String> tokens;
	private ArrayList<String> labels;
	private int paragraphNumber;
	private int sentenceNumber;
	
	public Sentence(String sentenceText, ArrayList<String> tokens, ArrayList<String> labels, int paragraphNumber, int sentenceNumber) {
		this.sentenceText = sentenceText;
		this.tokens = tokens;
		this.labels = labels;
		this.paragraphNumber = paragraphNumber;
		this.sentenceNumber = sentenceNumber;
	}

	public String getSentenceText() {
		return sentenceText;
	}

	public void setSentenceText(String sentenceText) {
		this.sentenceText = sentenceText;
	}

	public int getParagraphNumber() {
		return paragraphNumber;
	}

	public void setParagraphNumber(int paragraphNumber) {
		this.paragraphNumber = paragraphNumber;
	}

	public int getSentenceNumber() {
		return sentenceNumber;
	}

	public void setSentenceNumber(int sentenceNumber) {
		this.sentenceNumber = sentenceNumber;
	}

	public ArrayList<String> getTokens() {
		return tokens;
	}

	public void setTokens(ArrayList<String> tokens) {
		this.tokens = tokens;
	}

	public ArrayList<String> getLabels() {
		return labels;
	}

	public void setLabels(ArrayList<String> labels) {
		this.labels = labels;
	}
}

